# Copyright (C) 2017  Patrick Reichenberger (University of Passau) <patrick.reichenberger@t-online.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from scapy.all import *
from datetime import datetime

# Station Address (maybe better in databus)
station_addr = 0x1e28


# Structure of control field formats
class i_frame(Packet):
    name = "i_frame"
    fields_desc = [XByteField("Start", 0x68),
                   ByteField("LenAPDU", None),
                   LEShortField("SendSeq", 0x0),
                   LEShortField("RecvSeq", 0x0)]

    # Compute length
    def post_build(self, p, pay):
        if self.LenAPDU is None:
            length = len(pay) + 4
            p = p[:1] + struct.pack("!B", length) + p[2:]
        return p + pay


class u_frame(Packet):
    name = "u_frame"
    fields_desc = [XByteField("Start", 0x68),
                   ByteField("LenAPDU", 0x04),
                   XByteField("Type", 0x07),
                   X3BytesField("Default", 0x000000)]


class s_frame(Packet):
    name = "s_frame"
    fields_desc = [XByteField("Start", 0x68),
                   ByteField("LenAPDU", 0x04),
                   XByteField("Type", 0x01),
                   XByteField("Default", 0x00),
                   LEShortField("RecvSeq", 0x0)]  # 0001 is in packet 01 00 and that's true, bec 1 is LSB


TypeIdentification = {
    # Single point information
    "M_SP_NA_1": 1,
    # Single point information with time tag
    "M_SP_TA_1": 2,
    # Double point information
    "M_DP_NA_1": 3,
    # Double point information with time tag
    "M_DP_TA_1": 4,
    # Step position information
    # "M_ST_NA_1": 5,
    # Step position information with time tag
    # "M_ST_TA_1": 6,
    # Bit string of 32 bit
    # "M_BO_NA_1": 7,
    # Bit string of 32 bit with time tag
    # "M_BO_TA_1": 8,
    # Measured value, normalized value
    # "M_ME_NA_1": 9,
    # Measured value, normalized value with time tag
    # "M_ME_TA_1": 10,
    # Measured value, scaled value
    "M_ME_NB_1": 11,
    # Measured value, scaled value with time tag
    "M_ME_TB_1": 12,
    # Measured value, short floating point value
    "M_ME_NC_1": 13,
    # Measured value, short floating point value with time tag
    "M_ME_TC_1": 14,
    # Integrated totals
    # "M_IT_NA_1": 15,
    # Integrated totals with time tag
    # "M_IT_TA_1": 16,
    # Event of protection equipment with time tag
    # "M_EP_TA_1": 17,
    # Packed start events of protection equipment with time tag
    # "M_EP_TB_1": 18,
    # Packed output circuit information of protection equipment with time tag
    # "M_EP_TC_1": 19,
    # Packed single-point information with status change detection
    # "M_PS_NA_1": 20,
    # Measured value, normalized value without quality descriptor
    # "M_ME_ND_1": 21,

    # Process telegrams with long time tag ( 7 octets ) :
    # Single point information with time tag CP56Time2a
    "M_SP_TB_1": 30,
    # Double point information with time tag CP56Time2a
    "M_DP_TB_1": 31,
    # Step position information with time tag CP56Time2a
    # "M_ST_TB_1": 32,
    # Bit string of 32 bit with time tag CP56Time2a
    # "M_BO_TB_1": 33,
    # Measured value, normalized value with time tag CP56Time2a
    # "M_ME_TD_1": 34,
    # Measured value, scaled value with time tag CP56Time2a
    "M_ME_TE_1": 35,
    # Measured value, short floating point value with time tag CP56Time2a
    "M_ME_TF_1": 36,
    # Integrated totals with time tag CP56Time2a
    # "M_IT_TB_1": 37,
    # Event of protection equipment with time tag CP56Time2a
    # "M_EP_TD_1": 38,
    # Packed start events of protection equipment with time tag CP56time2a
    # "M_EP_TE_1": 39,
    # Packed output circuit information of protection equipment with time tag CP56Time2a
    # "M_EP_TF_1": 40,

    # Process information in control direction:
    # Single command
    "C_SC_NA_1": 45,
    # Double command
    "C_DC_NA_1": 46,
    # Regulating step command
    # "C_RC_NA_1": 47,
    # Setpoint command, normalized value
    # "C_SE_NA_1": 48,
    # Setpoint command, scaled value
    "C_SE_NB_1": 49,
    # Setpoint command, short floating point value
    "C_SE_NC_1": 50,
    # Bit string  32 bit
    # "C_BO_NA_1": 51,

    # Command telegrams with long time tag ( 7 octets ):
    # Single command with time tag CP56Time2a
    # "C_SC_TA_1": 58,
    # Double command with time tag CP56Time2a
    # "C_DC_TA_1": 59,
    # Regulating step command with time tag CP56Time2a
    # "C_RC_TA_1": 60,
    # Setpoint command, normalized value with time tag CP56Time2a
    # "C_SE_TA_1": 61,
    # Setpoint command, scaled value with time tag CP56Time2a
    # "C_SE_TB_1": 62,
    # Setpoint command, short floating point value with time tag CP56Time2a
    # "C_SE_TC_1": 63,
    # Bit string 32 bit with time tag CP56Time2a
    # "C_BO_TA_1": 64,

    # System information  in monitoring direction :
    # End of initialization
    # "M_EI_NA_1": 70,

    # System information in control direction :
    # (General-) Interrogation command
    "C_IC_NA_1": 100,
    # Counter interrogation command
    # "C_CI_NA_1": 101,
    # Read command
    # "C_RD_NA_1": 102,
    # Clock synchronization command
    # "C_CS_NA_1": 103,
    # ( IEC 101 ) Test command
    # "C_TS_NB_1": 104,
    # Reset process command
    # "C_RP_NC_1": 105,
    # ( IEC 101 ) Delay acquisition command
    # "C_CD_NA_1": 106,
    # Test command with time tag CP56Time2a
    # "C_TS_TA_1": 107,

    #  Parameter in control direction :
    # Parameter of measured value, normalized value
    # "P_ME_NA_1": 110,
    # Parameter of measured value, scaled value
    # "P_ME_NB_1": 111,
    # Parameter of measured value, short floating point value
    # "P_ME_NC_1": 112,
    # Parameter activation
    # "P_AC_NA_1": 113,

    # File transfer:
    # File ready
    # "F_FR_NA_1": 120,
    # Section ready
    # "F_SR_NA_1": 121,
    # Call directory, select file, call file, call section
    # "F_SC_NA_1": 122,
    # Last section, last segment
    # "F_LS_NA_1": 123,
    # Ack file, Ack section
    # "F_AF_NA_1": 124,
    # Segment
    # "F_SG_NA_1": 125,
    # Directory
    # "F_DR_TA_1": 126,
    # QueryLog - Request archive file
    # "F_SC_NB_1": 127
}


class LESignedShortField(Field):
    def __init__(self, name, default):
        Field.__init__(self, name, default, "<h")


class NormValueField(LESignedShortField):
    def i2repr(self, pkt, x):
        normalized = 2 * ((x+2**15)/((2**15+2**15.)-1)) - 1
        return self.i2h(pkt, normalized)


class CP56Time(Packet):
    name = "CP56Time"
    fields_desc = [
        ShortField("Ms", 0x0000),
        ByteField("Min", 0x00),
        ByteField("Hour", 0x0),
        ByteField("Day", 0x01),
        ByteField("Month", 0x01),
        ByteField("Year", 0x5b)]


class CP24Time(Packet):
    name = "CP24Time"
    fields_desc = [
        ShortField("Ms", 0x0000),
        ByteField("Min", 0x00)]

    def extract_padding(self, p):
        return "", p


class CP16Time(Packet):
    name = "CP24Time"
    fields_desc = [
        ShortField("Ms", 0x0000)]

    def extract_padding(self, p):
        return "", p


# Info Elements

class IOA(Packet):
    name = "IOA"
    fields_desc = [X3BytesField("IOA", 0x010000)]


class QOS(Packet):
    #  Quality of set-point command
    name = "QOS"
    fields_desc = [
        XBitField("S/E", 0, 1), XBitField("QL", 0, 7)]


class QDS(Packet):
    #  Quality descriptor
    name = "QDS"
    fields_desc = [
        XBitField("IV", 0, 1), XBitField("NT", 0, 1), XBitField("SB", 0, 1), XBitField("BL", 0, 1),
        XBitField("Padding", 0, 3), XBitField("OV", 0, 1)]

    def extract_padding(self, p):
        return "", p


class QDP(Packet):
    name = "QDP"
    fields_desc = [
        XBitField("IV", 0, 1), XBitField("NT", 0, 1), XBitField("SB", 0, 1), XBitField("BL", 0, 1),
        XBitField("EI", 0, 1), XBitField("Padding", 0, 3)]

    def extract_padding(self, p):
        return "", p


class SIQ(Packet):
    name = "SIQ"
    fields_desc = [
        #  XByteField("SIQ", 0x00)]
        #  Exacter representation for SIQ:
        XBitField("IV", 0, 1), XBitField("NT", 0, 1), XBitField("SB", 0, 1), XBitField("BL", 0, 1),
        XBitField("Padding", 0, 3), XBitField("SPI", 0, 1)]


class BSI(Packet):
    name = "BSI"
    fields_desc = [
        LEIntField("BSI", 0)]


class DIQ(Packet):
    name = "DIQ"
    fields_desc = [
        #  XByteField("DIQ", 0x00)]
        #  Exacter representation for DIQ:
        XBitField("IV", 0, 1), XBitField("NT", 0, 1), XBitField("SB", 0, 1), XBitField("BL", 0, 1),
        XBitField("Padding", 0, 2), XBitField("DPI", 0, 2)]


class VTI(Packet):
    name = "VTI"
    fields_desc = [
        XBitField("T", 0, 1), XBitField("Value", 0, 7)]


class NVA(Packet):
    # Normalized value
    name = "NVA"
    fields_desc = [
        NormValueField("NVA", 0x5000)]


class SVA(Packet):
    # Scaled value
    name = "SVA"
    fields_desc = [
        LESignedShortField("SVA", 0x50)]


class BCR(Packet):
    #  Binary Counter Reading
    name = "BCR"
    fields_desc = [
        LESignedIntField("Value", 0x0),
        XBitField("IV", 0, 1), XBitField("CA", 0, 1), XBitField("CY", 0, 1), XBitField("SeqNr", 0, 5)]


class SEP(Packet):
    name = "SEP"
    fields_desc = [
        XBitField("IV", 0, 1), XBitField("NT", 0, 1), XBitField("SB", 0, 1), XBitField("BL", 0, 1),
        XBitField("EI", 0, 1), XBitField("Padding", 0, 1), XBitField("ES", 0, 2)]

    def extract_padding(self, p):
        return "", p


class SPE(Packet):
    name = "SPE"
    fields_desc = [
        XBitField("Padding", 0, 2), XBitField("SRD", 0, 1), XBitField("SIE", 0, 1),
        XBitField("SL3", 0, 1), XBitField("SL2", 0, 1), XBitField("SL2", 0, 1), XBitField("GS", 0, 1)]

    def extract_padding(self, p):
        return "", p


class OCI(Packet):
    name = "OCI"
    fields_desc = [
        XBitField("Padding", 0, 4),
        XBitField("CL3", 0, 1), XBitField("CL2", 0, 1), XBitField("CL1", 0, 1), XBitField("GC", 0, 1)]


class SCD(Packet):
    name = "SCD"
    fields_desc = [
        LEShortField("Status", 0x0),  # LE?
        LEShortField("StatChaDet", 0x0)]  # LE?


class FloatField(Field):
    def __init__(self, name, default):
        Field.__init__(self, name, default, "<f")


# ASDU packets
class asdu_infobj_1(Packet):
    name = "M_SP_NA_1"
    fields_desc = [
        IOA,
        # SIQ]
        PacketField("SIQ", SIQ(), SIQ)]


class asdu_infobj_2(Packet):
    name = "M_SP_TA_1"
    fields_desc = [
        IOA,
        PacketField("SIQ", SIQ(), SIQ),
        PacketField("CP24Time", CP24Time(), CP24Time)]


class asdu_infobj_3(Packet):
    name = "M_DP_NA_1"
    fields_desc = [
        IOA,
        PacketField("DIQ", DIQ(), DIQ)]


class asdu_infobj_4(Packet):
    name = "M_DP_TA_1"
    fields_desc = [
        IOA,
        PacketField("DIQ", DIQ(), DIQ),
        PacketField("CP24Time", CP24Time(), CP24Time)]


class asdu_infobj_5(Packet):
    name = "M_ST_NA_1"
    fields_desc = [
        IOA,
        PacketField("VTI", VTI(), VTI),
        PacketField("QDS", QDS(), QDS)]


class asdu_infobj_6(Packet):
    name = "M_ST_TA_1"
    fields_desc = [
        IOA,
        PacketField("VTI", VTI(), VTI),
        PacketField("QDS", QDS(), QDS),
        PacketField("CP24Time", CP24Time(), CP24Time)]


class asdu_infobj_7(Packet):
    name = "M_BO_NA_1"
    fields_desc = [
        IOA,
        BSI,
        PacketField("QDS", QDS(), QDS)]


class asdu_infobj_8(Packet):
    name = "M_BO_TA_1"
    fields_desc = [
        IOA,
        BSI,
        PacketField("QDS", QDS(), QDS),
        PacketField("CP24Time", CP24Time(), CP24Time)]


class asdu_infobj_9(Packet):
    name = "M_ME_NA_1"
    fields_desc = [
        IOA,
        NVA,
        PacketField("QDS", QDS(), QDS)]


class asdu_infobj_10(Packet):
    name = "M_ME_TA_1"
    fields_desc = [
        IOA,
        NVA,
        PacketField("QDS", QDS(), QDS),
        PacketField("CP24Time", CP24Time(), CP24Time)]


class asdu_infobj_11(Packet):
    name = "M_ME_NB_1"
    fields_desc = [
        IOA,
        SVA,
        PacketField("QDS", QDS(), QDS)]


class asdu_infobj_12(Packet):
    name = "M_ME_TB_1"
    fields_desc = [
        IOA,
        SVA,
        PacketField("QDS", QDS(), QDS),
        PacketField("CP24Time", CP24Time(), CP24Time)]


class asdu_infobj_13(Packet):
    name = "M_ME_NC_1"
    fields_desc = [
        IOA,
        FloatField("FPNumber", 1),
        PacketField("QDS", QDS(), QDS)]


class asdu_infobj_14(Packet):
    name = "M_ME_TC_1"
    fields_desc = [
        IOA,
        IEEEFloatField("FPNumber", 0),
        PacketField("QDS", QDS(), QDS),
        PacketField("CP24Time", CP24Time(), CP24Time)]


class asdu_infobj_15(Packet):
    name = "M_IT_NA_1"
    fields_desc = [
        IOA,
        PacketField("BCR", BCR(), BCR)]


class asdu_infobj_16(Packet):
    name = "M_IT_TA_1"
    fields_desc = [
        IOA,
        PacketField("BCR", BCR(), BCR),
        PacketField("CP24Time", CP24Time(), CP24Time)]


class asdu_infobj_17(Packet):
    name = "M_EP_TA_1"
    fields_desc = [
        IOA,
        PacketField("SEP", SEP(), SEP),
        CP16Time,  # elapsed time
        PacketField("CP24Time", CP24Time(), CP24Time)]  # binary time


class asdu_infobj_18(Packet):
    name = "M_EP_TB_1"
    fields_desc = [
        IOA,
        PacketField("SPE", SPE(), SPE),
        PacketField("QDP", QDP(), QDP),
        CP16Time,  # elapsed time
        PacketField("CP24Time", CP24Time(), CP24Time)]  # binary time


class asdu_infobj_19(Packet):
    name = "M_EP_TC_1"
    fields_desc = [
        IOA,
        PacketField("OCI", OCI(), OCI),
        PacketField("QDP", QDP(), QDP),
        CP16Time,  # relay duration time
        PacketField("CP24Time", CP24Time(), CP24Time)]  # binary time


class asdu_infobj_20(Packet):
    name = "M_PS_NA_1"
    fields_desc = [
        IOA,
        PacketField("SCD", SCD(), SCD),
        PacketField("QDS", QDS(), QDS)]


class asdu_infobj_21(Packet):
    name = "M_ME_ND_1"
    fields_desc = [
        IOA,
        NVA]


class asdu_infobj_30(Packet):
    name = "M_SP_TB_1"
    fields_desc = [
        IOA,
        PacketField("SIQ", SIQ(), SIQ),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_31(Packet):
    name = "M_DP_TB_1"
    fields_desc = [
        IOA,
        PacketField("DIQ", DIQ(), DIQ),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_32(Packet):
    name = "M_ST_TA_1"
    fields_desc = [
        IOA,
        PacketField("VTI", VTI(), VTI),
        PacketField("QDS", QDS(), QDS),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_33(Packet):
    name = "M_BO_TB_1"
    fields_desc = [
        IOA,
        PacketField("BSI", BSI(), BSI),
        PacketField("QDS", QDS(), QDS),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_34(Packet):
    name = "M_ME_TD_1"
    fields_desc = [
        IOA,
        NVA,
        PacketField("QDS", QDS(), QDS),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_35(Packet):
    name = "M_ME_TE_1"
    fields_desc = [
        IOA,
        SVA,
        PacketField("QDS", QDS(), QDS),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_36(Packet):
    name = "M_ME_TF_1"
    fields_desc = [
        IOA,
        IEEEFloatField("FPNumber", 0),
        PacketField("QDS", QDS(), QDS),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_37(Packet):
    name = "M_IT_TB_1"
    fields_desc = [
        IOA,
        PacketField("BCR", BCR(), BCR),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_38(Packet):
    name = "M_EP_TD_1"
    fields_desc = [
        IOA,
        PacketField("SEP", SEP(), SEP),
        CP16Time,  # elapsed time
        PacketField("CP56Time", CP56Time(), CP56Time)]  # binary time


class asdu_infobj_39(Packet):
    name = "M_EP_TE_1"
    fields_desc = [
        IOA,
        PacketField("SPE", SPE(), SPE),
        PacketField("QDP", QDP(), QDP),
        CP16Time,  # relay duration time
        PacketField("CP56Time", CP56Time(), CP56Time)]  # binary time


class asdu_infobj_40(Packet):
    name = "M_EP_TF_1"
    fields_desc = [
        IOA,
        PacketField("OCI", OCI(), OCI),
        PacketField("QDP", QDP(), QDP),
        CP16Time,  # relay duration time
        PacketField("CP56Time", CP56Time(), CP56Time)]  # binary time


class asdu_infobj_45(Packet):
    name = "C_SC_NA_1"
    fields_desc = [
        IOA,
        #  XByteField("SCO", 0x00)]
        #  Exacter representation(2) for SCO:
        #  XBitField("S/E", 0, 1), XBitField("QU", 0, 5), XBitField("Padding", 0, 1), XBitField("SCS", 0, 1)]
        XBitField("QOC", 0, 6), XBitField("Padding", 0, 1), BitField("SCS", 0, 1)]


class asdu_infobj_46(Packet):
    name = "C_DC_NA_1"
    fields_desc = [
        IOA,
        #  XByteField("SCO", 0x00)]
        #  Exacter representation(2) for DCO:
        #  XBitField("S/E", 0, 1), XBitField("QU", 0, 5), XBitField("DCS", 0, 2)]
        XBitField("QOC", 0, 6), XBitField("DCS", 0, 2)]


class asdu_infobj_47(Packet):
    name = "C_RC_NA_1"
    fields_desc = [
        IOA,
        #  XByteField("SCO", 0x00)]
        #  Exacter representation(2) for RCO:
        #  XBitField("S/E", 0, 1), XBitField("QU", 0, 5), XBitField("RCS", 0, 2)]
        XBitField("QOC", 0, 6), XBitField("RCS", 0, 2)]


class asdu_infobj_48(Packet):
    name = "C_SE_NA_1"
    fields_desc = [
        IOA,
        # Normalized value
        NVA,
        PacketField("QOS", QOS(), QOS)]


class asdu_infobj_49(Packet):
    name = "C_SE_NB_1"
    fields_desc = [
        IOA,
        # Scaled value
        SVA,
        PacketField("QOS", QOS(), QOS)]


class asdu_infobj_50(Packet):
    name = "C_SE_NC_1"
    fields_desc = [
        IOA,
        IEEEFloatField("FPNumber", 0),
        PacketField("QOS", QOS(), QOS)]


class asdu_infobj_51(Packet):
    name = "C_BO_NA_1"
    fields_desc = [
        IOA,
        BSI]


# maybe in handle client
def calctime():
    currenttime = datetime.now()
    milliseconds = currenttime.microsecond / 1000
    seconds = currenttime.second
    ms = seconds * 1000 + milliseconds
    minutes = currenttime.minute
    hour = currenttime.hour
    day = currenttime.day
    month = currenttime.month
    year = currenttime.year
    cp56time = CP56Time()
    cp56time.setfieldval("Ms", ms)
    cp56time.setfieldval("Min", minutes)
    cp56time.setfieldval("Hour", hour)
    cp56time.setfieldval("Day", day)
    cp56time.setfieldval("Month", month)
    cp56time.setfieldval("Year", year)
    return cp56time


class asdu_infobj_58(Packet):
    name = "C_SC_TA_1"
    fields_desc = [
        IOA,
        #  XByteField("SCO", 0x00)]
        #  Exacter representation(2) for SCO:
        #  XBitField("S/E", 0, 1), XBitField("QU", 0, 5), XBitField("Padding", 0, 1), XBitField("SCS", 0, 1)]
        XBitField("QOC", 0, 6), XBitField("Padding", 0, 1), BitField("SCS", 0, 1),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_59(Packet):
    name = "C_DC_TA_1"
    fields_desc = [
        IOA,
        #  XByteField("DCO", 0x00)]
        #  Exacter representation(2) for DCO:
        #  XBitField("S/E", 0, 1), XBitField("QU", 0, 5), XBitField("DCS", 0, 2)]
        XBitField("QOC", 0, 6), XBitField("DCS", 0, 2),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_60(Packet):
    name = "C_RC_TA_1"
    fields_desc = [
        IOA,
        #  XByteField("RCO", 0x00)]
        #  Exacter representation(2) for RCO:
        #  XBitField("S/E", 0, 1), XBitField("QU", 0, 5), XBitField("RCS", 0, 2)]
        XBitField("QOC", 0, 6), XBitField("RCS", 0, 2),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_61(Packet):
    name = "C_SE_TA_1"
    fields_desc = [
        IOA,
        # Normalized value
        NVA,
        PacketField("QOS", QOS(), QOS),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_62(Packet):
    name = "C_SE_TB_1"
    fields_desc = [
        IOA,
        # Scaled value
        SVA,
        PacketField("QOS", QOS(), QOS),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_63(Packet):
    name = "C_SE_TC_1"
    fields_desc = [
        IOA,
        IEEEFloatField("FPNumber", 0),
        PacketField("QOS", QOS(), QOS),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_64(Packet):
    name = "C_BO_TA_1"
    fields_desc = [
        IOA,
        BSI,
        PacketField("QOS", QOS(), QOS),
        PacketField("CP56Time", CP56Time(), CP56Time)]


class asdu_infobj_100(Packet):
    name = "C_IC_NA_1"
    fields_desc = [
        X3BytesField("IOA", 0x0),
        ByteField("QOI", 0x14)]


class asdu_infobj_101(Packet):
    name = "C_CI_NA_1"
    fields_desc = [
        X3BytesField("IOA", 0x0),
        ByteField("QCC", 0x05)]


class asdu_infobj_102(Packet):
    name = "C_RD_NA_1"
    fields_desc = [
        X3BytesField("IOA", 0x0)]


class asdu_infobj_103(Packet):
    name = "C_CS_NA_1"
    fields_desc = [
        X3BytesField("IOA", 0x0),
        PacketField("CP56Time", CP56Time(), CP56Time)]


# IEC104 asdu head
class asdu_head(Packet):
    name = "asdu_head"
    fields_desc = [ByteField("TypeID", 0x05),  # Command Type
                   #  Exacter representation for variable structure qualifier
                   BitField("SQ", 0b0, 1), BitField("NoO", 1, 7),  # SQ and Number of Object
                   #  XByteField("NoO", 0x01),
                   #  Exacter representation for Cause of Transmission:
                   BitField("T", 0, 1), BitField("PN", 0, 1), BitField("COT", 6, 6),
                   # XByteField("COT", 0x06),
                   XByteField("OrigAddr", 0x00),
                   LEShortField("Addr", station_addr)]

    def guess_payload_class(self, payload):
        if self.TypeID == 1:
            if self.SQ == 0:
                # List--bind_layers(asdu_infobj_1, asdu_infobj_1)  # if SQ = 0
                # List--bind_layers(SIQ, asdu_infobj_1)  # if SQ = 0
                bind_layers(asdu_infobj_1, asdu_infobj_1)  # if SQ = 0
                bind_layers(SIQ, Padding)  # if SQ = 0
            else:
                # List--bind_layers(SIQ, Padding)  # if SQ = 1
                bind_layers(SIQ, SIQ)  # if SQ = 1
            return asdu_infobj_1
        elif self.TypeID == 2:
            return asdu_infobj_2
        elif self.TypeID == 3:
            if self.SQ == 0:
                bind_layers(asdu_infobj_3, asdu_infobj_3)  # if SQ = 0
                bind_layers(DIQ, Padding)  # if SQ = 0
            else:
                bind_layers(DIQ, DIQ)  # if SQ = 1
            return asdu_infobj_3
        elif self.TypeID == 4:
            return asdu_infobj_4
        elif self.TypeID == 5:
            if self.SQ == 0:
                bind_layers(asdu_infobj_5, asdu_infobj_5)  # if SQ = 0
                bind_layers(QDS, Padding)  # if SQ = 0
            else:
                bind_layers(QDS, VTI)  # if SQ = 1
            return asdu_infobj_5
        elif self.TypeID == 6:
            return asdu_infobj_6
        elif self.TypeID == 7:
            if self.SQ == 0:
                bind_layers(asdu_infobj_7, asdu_infobj_7)  # if SQ = 0
                bind_layers(QDS, Padding)  # if SQ = 0
            else:
                bind_layers(QDS, BSI)  # if SQ = 1
            return asdu_infobj_7
        elif self.TypeID == 8:
            return asdu_infobj_8
        elif self.TypeID == 9:
            if self.SQ == 0:
                bind_layers(asdu_infobj_9, asdu_infobj_9)  # if SQ = 0
                bind_layers(QDS, Padding)  # if SQ = 0
            else:
                bind_layers(QDS, NVA)  # if SQ = 1
            return asdu_infobj_9
        elif self.TypeID == 10:
            return asdu_infobj_10
        elif self.TypeID == 11:
            if self.SQ == 0:
                bind_layers(asdu_infobj_11, asdu_infobj_11)  # if SQ = 0
                bind_layers(QDS, Padding)  # if SQ = 0
            else:
                bind_layers(QDS, SVA)  # if SQ = 1
            return asdu_infobj_11
        elif self.TypeID == 12:
            return asdu_infobj_12
        elif self.TypeID == 13:
            if self.SQ == 0:
                bind_layers(asdu_infobj_13, asdu_infobj_13)  # if SQ = 0
                bind_layers(QDS, Padding)  # if SQ = 0
            else:
                bind_layers(QDS, FloatField)  # if SQ = 1
            return asdu_infobj_13
        elif self.TypeID == 14:
            return asdu_infobj_14
        elif self.TypeID == 15:
            return asdu_infobj_15
        elif self.TypeID == 16:
            return asdu_infobj_16
        elif self.TypeID == 17:
            return asdu_infobj_17
        elif self.TypeID == 18:
            return asdu_infobj_18
        elif self.TypeID == 19:
            return asdu_infobj_19
        elif self.TypeID == 20:
            return asdu_infobj_20
        elif self.TypeID == 21:
            return asdu_infobj_21
        elif self.TypeID == 30:
            return asdu_infobj_30
        elif self.TypeID == 31:
            return asdu_infobj_31
        elif self.TypeID == 32:
            return asdu_infobj_32
        elif self.TypeID == 33:
            return asdu_infobj_33
        elif self.TypeID == 34:
            return asdu_infobj_34
        elif self.TypeID == 35:
            return asdu_infobj_35
        elif self.TypeID == 36:
            return asdu_infobj_36
        elif self.TypeID == 37:
            return asdu_infobj_37
        elif self.TypeID == 38:
            return asdu_infobj_38
        elif self.TypeID == 39:
            return asdu_infobj_39
        elif self.TypeID == 40:
            return asdu_infobj_40
        elif self.TypeID == 45:
            return asdu_infobj_45
        elif self.TypeID == 46:
            return asdu_infobj_46
        elif self.TypeID == 47:
            return asdu_infobj_47
        elif self.TypeID == 48:
            return asdu_infobj_48
        elif self.TypeID == 49:
            return asdu_infobj_49
        elif self.TypeID == 50:
            return asdu_infobj_50
        elif self.TypeID == 51:
            return asdu_infobj_51
        elif self.TypeID == 58:
            return asdu_infobj_58
        elif self.TypeID == 59:
            return asdu_infobj_59
        elif self.TypeID == 60:
            return asdu_infobj_60
        elif self.TypeID == 61:
            return asdu_infobj_61
        elif self.TypeID == 62:
            return asdu_infobj_62
        elif self.TypeID == 63:
            return asdu_infobj_63
        elif self.TypeID == 64:
            return asdu_infobj_64
        # elif self.TypeID == 70:
        #    return asdu_infobj_70
        elif self.TypeID == 100:
            return asdu_infobj_100
        elif self.TypeID == 101:
            return asdu_infobj_101
        elif self.TypeID == 102:
            return asdu_infobj_102
        elif self.TypeID == 103:
            return asdu_infobj_103


bind_layers(i_frame, asdu_head)
bind_layers(asdu_head, asdu_infobj_1, {'TypeID': 1})
bind_layers(asdu_head, asdu_infobj_2, {'TypeID': 2})
bind_layers(asdu_head, asdu_infobj_3, {'TypeID': 3})
bind_layers(asdu_head, asdu_infobj_4, {'TypeID': 4})
bind_layers(asdu_head, asdu_infobj_5, {'TypeID': 5})
bind_layers(asdu_head, asdu_infobj_6, {'TypeID': 6})
bind_layers(asdu_head, asdu_infobj_7, {'TypeID': 7})
bind_layers(asdu_head, asdu_infobj_8, {'TypeID': 8})
bind_layers(asdu_head, asdu_infobj_9, {'TypeID': 9})
bind_layers(asdu_head, asdu_infobj_10, {'TypeID': 10})
bind_layers(asdu_head, asdu_infobj_11, {'TypeID': 11})
bind_layers(asdu_head, asdu_infobj_12, {'TypeID': 12})
bind_layers(asdu_head, asdu_infobj_13, {'TypeID': 13})
bind_layers(asdu_head, asdu_infobj_14, {'TypeID': 14})
bind_layers(asdu_head, asdu_infobj_15, {'TypeID': 15})
bind_layers(asdu_head, asdu_infobj_16, {'TypeID': 16})
bind_layers(asdu_head, asdu_infobj_17, {'TypeID': 17})
bind_layers(asdu_head, asdu_infobj_18, {'TypeID': 18})
bind_layers(asdu_head, asdu_infobj_19, {'TypeID': 19})
bind_layers(asdu_head, asdu_infobj_20, {'TypeID': 20})
bind_layers(asdu_head, asdu_infobj_21, {'TypeID': 21})

bind_layers(asdu_head, asdu_infobj_30, {'TypeID': 30})
bind_layers(asdu_head, asdu_infobj_31, {'TypeID': 31})
bind_layers(asdu_head, asdu_infobj_32, {'TypeID': 32})
bind_layers(asdu_head, asdu_infobj_33, {'TypeID': 33})
bind_layers(asdu_head, asdu_infobj_34, {'TypeID': 34})
bind_layers(asdu_head, asdu_infobj_35, {'TypeID': 35})
bind_layers(asdu_head, asdu_infobj_36, {'TypeID': 36})
bind_layers(asdu_head, asdu_infobj_37, {'TypeID': 37})
bind_layers(asdu_head, asdu_infobj_38, {'TypeID': 38})
bind_layers(asdu_head, asdu_infobj_39, {'TypeID': 39})
bind_layers(asdu_head, asdu_infobj_40, {'TypeID': 40})

bind_layers(asdu_head, asdu_infobj_45, {'TypeID': 45})
bind_layers(asdu_head, asdu_infobj_46, {'TypeID': 46})
bind_layers(asdu_head, asdu_infobj_47, {'TypeID': 47})
bind_layers(asdu_head, asdu_infobj_48, {'TypeID': 48})
bind_layers(asdu_head, asdu_infobj_49, {'TypeID': 49})
bind_layers(asdu_head, asdu_infobj_50, {'TypeID': 50})
bind_layers(asdu_head, asdu_infobj_51, {'TypeID': 51})

bind_layers(asdu_head, asdu_infobj_58, {'TypeID': 58})
bind_layers(asdu_head, asdu_infobj_59, {'TypeID': 59})
bind_layers(asdu_head, asdu_infobj_60, {'TypeID': 60})
bind_layers(asdu_head, asdu_infobj_61, {'TypeID': 61})
bind_layers(asdu_head, asdu_infobj_62, {'TypeID': 62})
bind_layers(asdu_head, asdu_infobj_63, {'TypeID': 63})
bind_layers(asdu_head, asdu_infobj_64, {'TypeID': 64})

bind_layers(asdu_head, asdu_infobj_100, {'TypeID': 100})
bind_layers(asdu_head, asdu_infobj_101, {'TypeID': 101})
bind_layers(asdu_head, asdu_infobj_102, {'TypeID': 102})
bind_layers(asdu_head, asdu_infobj_103, {'TypeID': 103})

# For SQ=1 and SQ=0, experimental..
# bind_layers(asdu_infobj_1, asdu_infobj_1)
# bind_layers(asdu_infobj_1, SIQ)
# bind_layers(SIQ, SIQ)

# bind_layers(asdu_infobj_1, asdu_infobj_1, {'SQ': 0})
# bind_layers(SIQ, asdu_infobj_1, {'SQ': 0})

bind_layers(SIQ, Padding)
bind_layers(DIQ, Padding)
bind_layers(VTI, Padding)
bind_layers(QDS, Padding)
bind_layers(BCR, Padding)
bind_layers(SEP, Padding)
bind_layers(SPE, Padding)
bind_layers(QDP, Padding)
bind_layers(QOS, Padding)

STARTDT_act = u_frame(Type=0x07)
STARTDT_con = u_frame(Type=0x0B)
STOPDT_act = u_frame(Type=0x13)
STOPDT_con = u_frame(Type=0x23)
TESTFR_act = u_frame(Type=0x43)
TESTFR_con = u_frame(Type=0x83)

u_list = {'0x7': 'STARTDT_ACT',
          '0xB': 'STARTDT_CON',
          '0x13': 'STOPDT_act',
          '0x23': 'STOPDT_con',
          '0x43': 'TESTFR_act',
          '0x83': 'TESTFR_con'}

# ==== Timeouts ==== old.....
# Timeout of connection establishment
# T_0 = 30
# Timeout of send or test APDUs (Wartezeit auf Quittung)
# T_1 = 15
# Timeout for acknowledges in case of no data messages T_2 < T_1 (Quittieren nach x sek)
# T_2 = 10
# Timeout for sending test frames in case of a long idle state
# T_3 = 21

# ==== Other parameters ====
# Maximum difference receive sequence number to send state variable (Max. Anzahl unquittierter Telegramme)
# k = 12
# Latest acknowledge after receiving w I-format APDUs (Quittieren nach w Telegrammen)
# w = 8
# Maximum frame size (in bytes)
# MaxFrameSize = 254

# Testing a packet:

# test2 = i_frame() / asdu_head()
#  test2.SIQ = SIQ(IV=1)  #  Change value in PacketField-packet
# hexdump(test2)
# test2.show()



# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Test core features for Conpot's virtual file system
"""
import conpot.core as conpot_core
from conpot.core.filesystem import SubAbstractFS
import unittest
import conpot
from freezegun import freeze_time
import fs
from datetime import datetime
from fs import permissions
from fs.time import epoch_to_datetime


class TestFileSystem(unittest.TestCase):
    """
        Tests related to Conpot's virtual file system.
    """

    def setUp(self):
        conpot_core.initialize_vfs()
        self.test_vfs = conpot_core.get_vfs()

    def tearDown(self):
        self.test_vfs.close()

    def test_listdir(self):
        _list = self.test_vfs.listdir('.')
        self.assertIn('data', _list)

    def test_chmod(self):
        # let us first create some directories!
        # TODO: check file create permissions and directory create permissions
        self.test_vfs.chmod('/data', 0o500, recursive=True)
        # Test that changes are reflected in the FS
        self.assertEqual(fs.permissions.Permissions.create(0o500),
                         self.test_vfs.getinfo(path='/data', namespaces=['access']).permissions)
        # No changes made in the actual file system.
        self.assertNotEqual(fs.permissions.Permissions.create(0o500),
                            self.test_vfs.getinfo(path='/data', get_actual=True, namespaces=['access']).permissions)

    def test_chown(self):
        self.test_vfs.register_user(name='test_user', uid=3000)
        self.test_vfs.create_group(name='test_grp', gid=2000)
        # do chown
        self.test_vfs.chown('/data', uid=3000, gid=2000)
        # check uid
        self.assertEqual(self.test_vfs.getinfo('/data', namespaces=['access']).uid, 3000)
        # actual uid shouldn't have changed
        self.assertNotEqual(self.test_vfs.getinfo('/data', get_actual=True, namespaces=['access']).uid, 3000)
        # check gid
        self.assertEqual(self.test_vfs.getinfo('/data', namespaces=['access']).gid, 2000)
        # FIXME: self.assertNotEqual(self.test_vfs.getinfo('/data', get_actual=True, namespaces=['access']).gid, 2000)
        # check file username
        self.assertEqual(self.test_vfs.getinfo('/data', namespaces=['access']).user, 'test_user')
        self.assertNotEqual(self.test_vfs.getinfo('/data', get_actual=True, namespaces=['access']).user, 'test_user')
        # check file group
        self.assertEqual(self.test_vfs.getinfo('/data', namespaces=['access']).group, 'test_grp')
        self.assertNotEqual(self.test_vfs.getinfo('/data', get_actual=True, namespaces=['access']).group, 'test_grp')
        # TODO: check for exceptions when user does not exist in the user/grp tables.

    def test_jail(self):
        """Test for checking chroot jail a subfilesystem"""
        # TODO: check for fs.error.permission denied error when we do a '../'
        self.assertTrue(self.test_vfs.create_jail('/data'))

    def test_mkdir(self):
        self.test_vfs.makedir('/dummy_dir', permissions=0o500)
        self.assertTrue(self.test_vfs.norm_path('/dummy_dir'))
        self.assertNotEqual(self.test_vfs.getinfo('/dummy_dir', get_actual=True, namespaces=['access']).permissions,
                            fs.permissions.Permissions.create(0o500))
        # check the usr/grp that created the folder
        self.assertEqual(self.test_vfs.getinfo('/dummy_dir', namespaces=['access']).uid, self.test_vfs.default_uid)

    def test_open_dir(self):
        self.assertIsInstance(self.test_vfs.opendir('/data'), SubAbstractFS)

    def test_get_cwd(self):
        self.assertEqual('/', self.test_vfs.getcwd())

    @freeze_time("2018-07-15 17:51:17", tz_offset=-4)
    def test_snapshot(self):
        self.assertEqual({'date-time', 'snapshot-data'}, set(self.test_vfs.take_snapshot().keys()))
        self.assertEqual(self.test_vfs.take_snapshot()['date-time'], datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

    @freeze_time("2018-07-15 17:51:17", tz_offset=-4)
    def test_stat(self):
        # TODO: Fix these to appropriate values -- check if the values are valid or not.
        # check if all relevant attributes exist in stat
        self.assertTrue(all(
            True if i in set(self.test_vfs.stat('/data').keys()) else False
            for i in {'st_uid', 'st_gid', 'st_mode', 'st_atime', 'st_mtime', 'st_mtime_ns', 'st_ctime_ns', 'st_ctime'}
            ))

    @freeze_time("2018-07-15 17:51:17", tz_offset=-4)
    def test_getmtime(self):
        _mtime = self.test_vfs.getmtime('/data')
        self.assertFalse(_mtime == datetime.now())

    def test_get_permissions(self):
        self.assertEqual(self.test_vfs.getinfo('/data', namespaces=['access']).permissions.as_str(),
                         self.test_vfs.get_permissions('/data'))

    def test_remove(self):
        self.test_vfs.touch('test_remove.txt')
        self.test_vfs.setinfo('test_remove.txt', {})
        self.test_vfs.remove('test_remove.txt')
        self.assertNotIn('test_remove.txt', self.test_vfs._cache.keys())

    def test_removedir(self):
        self.test_vfs.makedir('/dummy')
        self.test_vfs.removedir('/dummy')
        self.assertNotIn('/dummy', self.test_vfs._cache.keys())

    def test_readlink(self):
        self.assertIsNone(self.test_vfs.readlink('/data'))

    def test_mkdirs(self):
        self.test_vfs.makedirs('demo/demo')
        self.assertTrue(self.test_vfs.get_permissions('/demo/demo'))

    def test_openbin_file(self):
        with self.test_vfs.openbin('new_file', mode='wb') as _file:
            _file.write(b'This is just a test')
        self.assertIn('new_file', self.test_vfs.listdir('/'))
        _test = self.test_vfs.gettext('/new_file')
        self.test_vfs.getinfo('new_file', namespaces=['basic'])
        self.assertEqual(_test, 'This is just a test')

    def test_open_file(self):
        with self.test_vfs.open('new_file', mode='w+') as _file:
            _file.write('This is just a test')
        self.assertIn('new_file', self.test_vfs.listdir('/'))
        _test = self.test_vfs.gettext('/new_file')
        self.test_vfs.getinfo('new_file', namespaces=['basic'])
        self.assertEqual(_test, 'This is just a test')

    @freeze_time("2018-07-15 17:51:17")
    def test_format_list(self):
        self.test_vfs.settimes('/data', accessed=datetime.now(), modified=datetime.now())
        self._f_list = self.test_vfs.format_list('/', self.test_vfs.listdir('/'))
        [_result] = [i for i in self._f_list]
        self.assertIn('root', _result)
        self.assertIn('Jul 15 17:51', _result)

    @freeze_time("2028-07-15 17:51:17")
    def test_utime(self):
        self.test_vfs.utime('/data', accessed=datetime.now(), modified=datetime.now())
        self.assertEqual(self.test_vfs.getinfo('/data', namespaces=['details']).modified.ctime(),
                         datetime.now().ctime())
        self.assertEqual(self.test_vfs.getinfo('/data', namespaces=['details']).accessed.ctime(),
                         datetime.now().ctime())

    def test_access(self):
        # check it the root user has all the permissions
        self.assertTrue(self.test_vfs.access('/data', 0, required_perms='rwx'))
        # create some random group and check permissions for that
        self.test_vfs.create_group('random', 220)
        self.assertFalse(self.test_vfs.access('/data', 'random', required_perms='x'))

    @freeze_time("2028-07-15 17:51:17")
    def test_movedir(self):
        # move a directory - retain it's contents
        _uid = self.test_vfs.getinfo('/data', namespaces=['access']).uid
        _gid = self.test_vfs.getinfo('/data', namespaces=['access']).gid
        _perms = self.test_vfs.getinfo('/data', namespaces=['access']).permissions
        _user = self.test_vfs.getinfo('/data', namespaces=['access']).user
        _group = self.test_vfs.getinfo('/data', namespaces=['access']).group
        _accessed = self.test_vfs.getinfo('/data', namespaces=['details']).accessed
        _modified = self.test_vfs.getinfo('/data', namespaces=['details']).modified
        self.test_vfs.movedir('/data', '/data_move', create=True)
        self.assertEqual(self.test_vfs.getinfo('/data_move', namespaces=['access']).uid, _uid)
        self.assertEqual(self.test_vfs.getinfo('/data_move', namespaces=['access']).gid, _gid)
        self.assertEqual(self.test_vfs.getinfo('/data_move', namespaces=['access']).permissions, _perms)
        self.assertEqual(self.test_vfs.getinfo('/data_move', namespaces=['access']).user, _user)
        self.assertEqual(self.test_vfs.getinfo('/data_move', namespaces=['access']).group, _group)
        # accessed and modified file must not be the same.
        self.test_vfs.settimes('/data_move', accessed=datetime.now(), modified=datetime.now())
        self.assertNotEqual(self.test_vfs.getinfo('/data_move', namespaces=['details']).accessed, _accessed)
        self.assertNotEqual(self.test_vfs.getinfo('/data_move', namespaces=['details']).modified, _modified)

    @freeze_time("2028-07-15 17:51:17")
    def test_copydir(self):
        # copy a directory
        _uid = self.test_vfs.getinfo('/data', namespaces=['access']).uid
        _gid = self.test_vfs.getinfo('/data', namespaces=['access']).gid
        _perms = self.test_vfs.getinfo('/data', namespaces=['access']).permissions
        _user = self.test_vfs.getinfo('/data', namespaces=['access']).user
        _group = self.test_vfs.getinfo('/data', namespaces=['access']).group
        self.test_vfs.copydir('/data', '/data_move', create=True)
        self.assertEqual(self.test_vfs.getinfo('/data_move', namespaces=['access']).uid, _uid)
        self.assertEqual(self.test_vfs.getinfo('/data_move', namespaces=['access']).gid, _gid)
        self.assertEqual(self.test_vfs.getinfo('/data_move', namespaces=['access']).permissions, _perms)
        self.assertEqual(self.test_vfs.getinfo('/data_move', namespaces=['access']).user, _user)
        self.assertEqual(self.test_vfs.getinfo('/data_move', namespaces=['access']).group, _group)
        self.assertEqual(self.test_vfs.listdir('/'), ['data', 'data_move'])

    @freeze_time("2028-07-15 17:51:17")
    def test_copyfile(self):
        # create a copy of a file in a separate directory
        with self.test_vfs.open('test_fs.txt', mode='w+') as _file:
            _file.write('This is just a test file checking copyfile')
        self.test_vfs.copy(src_path='test_fs.txt', dst_path='test_fs_copy.txt', overwrite=True)
        _text = self.test_vfs.gettext('test_fs_copy.txt')
        self.assertEqual(_text, 'This is just a test file checking copyfile')
        self.assertTrue(self.test_vfs.getbasic('test_fs_copy.txt'))

    @freeze_time("2028-07-15 17:51:17")
    def test_movefile(self):
        # create a copy of a file in a separate directory
        with self.test_vfs.open('test_fs.txt', mode='w') as _file:
            _file.write('This is just a test file checking copyfile')
        _uid = self.test_vfs.getinfo('test_fs.txt', namespaces=['access']).uid
        self.test_vfs.move('test_fs.txt', 'test_fs_copy.txt', overwrite=True)
        _text = self.test_vfs.gettext('test_fs_copy.txt')
        self.assertEqual(self.test_vfs.getinfo('test_fs_copy.txt', namespaces=['access']).uid, _uid)
        self.assertEqual(_text, 'This is just a test file checking copyfile')


class TestSubFileSystem(unittest.TestCase):
    """
        Tests related to Conpot's virtual sub file system. This would test fs generated folders for each and
        every protocol.
    """

    def setUp(self):
        conpot_core.initialize_vfs()
        self._vfs = conpot_core.get_vfs()
        self._vfs.register_user('test_user', 13)
        self._vfs.create_group('test_grp', 13)
        self.test_vfs = self._vfs.mount_fs(
            fs_url='/'.join(conpot.__path__ + ['tests/data/test_data_fs/vfs']),
            dst_path='/data',
            owner_uid=13,
            group_gid=13,
            perms=0o750
        )

    def tearDown(self):
        self._vfs.close()

    def test_listdir(self):
        _list = self.test_vfs.listdir('.')
        self.assertIn('vfs.txt',  _list)

    def test_chmod(self):
        self.test_vfs.chmod('vfs.txt', 0o500, recursive=True)
        # Test that changes are reflected in the FS
        self.assertEqual(fs.permissions.Permissions.create(0o500),
                         self.test_vfs.getinfo(path='vfs.txt', namespaces=['access']).permissions)
        # No changes made in the actual file system.
        self.assertNotEqual(fs.permissions.Permissions.create(0o500),
                            self.test_vfs.getinfo(path='vfs.txt', get_actual=True, namespaces=['access']).permissions)

    def test_chown(self):
        self.test_vfs.register_user(name='new_user', uid=3000)
        self.test_vfs.create_group(name='new_grp', gid=2000)
        # do chown
        self.test_vfs.chown('vfs.txt', uid=3000, gid=2000)
        # check uid
        self.assertEqual(self.test_vfs.getinfo('vfs.txt', namespaces=['access']).uid, 3000)
        # actual uid shouldn't have changed
        self.assertNotEqual(self.test_vfs.getinfo('vfs.txt', get_actual=True, namespaces=['access']).uid, 3000)

    def test_mkdir(self):
        self.test_vfs.makedir('/dummy', permissions=0o500)
        self.assertFalse(self.test_vfs.access('/dummy', self.test_vfs.default_uid, 'rwx'))
        self.test_vfs.removedir('/dummy')
        self.assertNotIn('/dummy', self.test_vfs._cache.keys())

    def test_mkdirs(self):
        self.test_vfs.makedirs('/demo/demo')
        self.assertTrue(self.test_vfs.get_permissions('/demo/demo'))

    @freeze_time("2018-07-17 17:51:17")
    def test_open_file(self):
        with self.test_vfs.open('new_file', mode='wb') as _file:
            _file.write(b'This is just a test')
        self.assertIn('new_file', self.test_vfs.listdir('/'))
        self.test_vfs.settimes('/new_file', accessed=datetime.now(), modified=datetime.now())
        _test = self.test_vfs.gettext('/new_file')
        self.assertEqual(_test, 'This is just a test')
        self.assertEqual(self.test_vfs.getinfo('new_file', namespaces=['details']).modified.ctime(),
                         datetime.now().ctime())

    def test_get_cwd(self):
        self.assertEqual(
            self.test_vfs.getcwd(), '/data'
        )

    @freeze_time("2028-07-15 17:51:17")
    def test_utime(self):
        self.test_vfs.utime('vfs.txt', accessed=datetime.now(), modified=datetime.now())
        self.assertEqual(self.test_vfs.getinfo('vfs.txt', namespaces=['details']).modified.ctime(),
                         datetime.now().ctime())
        self.assertEqual(self.test_vfs.getinfo('vfs.txt', namespaces=['details']).accessed.ctime(),
                         datetime.now().ctime())

    @freeze_time("2018-07-15 17:51:17", tz_offset=-4)
    def test_stat(self):
        # TODO: Fix these to appropriate values -- check if the values are valid or not.
        # check if all relevant attributes exist in stat
        self.assertTrue(all(
            True if i in set(self.test_vfs.stat('vfs.txt').keys()) else False
            for i in {'st_uid', 'st_gid', 'st_mode', 'st_atime', 'st_mtime', 'st_mtime_ns', 'st_ctime_ns', 'st_ctime'}
            ))

    def test_get_permissions(self):
        self.assertEqual(self.test_vfs.getinfo('vfs.txt', namespaces=['access']).permissions.as_str(),
                         self.test_vfs.get_permissions('vfs.txt'))

    @freeze_time("2018-07-15 17:51:17")
    def test_set_time(self):
        """Test for changing time in the file system."""
        self.test_vfs.settimes('vfs.txt', accessed=epoch_to_datetime(103336010), modified=epoch_to_datetime(103336010))
        self.assertEqual(self.test_vfs.getinfo('vfs.txt', namespaces=['details']).accessed,
                         epoch_to_datetime(103336010))
        self.assertEqual(self.test_vfs.getinfo('vfs.txt', namespaces=['details']).modified,
                         epoch_to_datetime(103336010))

    def test_access(self):
        # check it the root user has all the permissions
        self.assertFalse(self.test_vfs.access('/vfs.txt', 0, required_perms='rwx'))
        self.assertEqual(self.test_vfs.getinfo('/vfs.txt', namespaces=['access']).uid, 13)
        self.assertTrue(self.test_vfs.access('/vfs.txt', 13, required_perms='rwx'))
        # create some random group and check permissions for that
        self.test_vfs.create_group('random', 220)
        self.assertFalse(self.test_vfs.access('/vfs.txt', 'random', required_perms='x'))
        # create a new user called test_access and add it group with gid 13
        self.test_vfs.register_user('test_access', 45)
        self.test_vfs.add_users_to_group(13, [45])
        self.assertTrue(self.test_vfs.access('/vfs.txt', 45, required_perms='rx'))
        self.assertFalse(self.test_vfs.access('/vfs.txt', 45, required_perms='rwx'))

    def test_remove(self):
        self.test_vfs.touch('test_remove.txt')
        self.test_vfs.setinfo('test_remove.txt', {})
        self.test_vfs.remove('test_remove.txt')
        self.assertNotIn('test_remove.txt', self.test_vfs._cache.keys())

    def test_removedir(self):
        self.test_vfs.makedir('/dummy')
        self.test_vfs.removedir('/dummy')
        self.assertNotIn('/dummy', self.test_vfs._cache.keys())

    def test_readlink(self):
        # FIXME: add tests for a file that is actually a link!
        self.assertIsNone(self.test_vfs.readlink('vfs.txt'))

    @freeze_time("2018-07-15 17:51:17")
    def test_format_list(self):
        self.test_vfs.settimes('vfs.txt', accessed=datetime.now(), modified=datetime.now())
        self._f_list = self.test_vfs.format_list('/', self.test_vfs.listdir('/'))
        [_result] = [i for i in self._f_list]
        self.assertIn(self.test_vfs.default_user, _result)
        self.assertIn('Jul 15 17:51', _result)


if __name__ == '__main__':
    unittest.main()


# Copyright (C) 2014  Andrea De Pasquale <andrea@de-pasquale.name>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging

from . import commands


logger = logging.getLogger(__name__)


class CommandResponder(object):
    COMMAND_NOT_FOUND = (
        "\r\n"
        "? Command not found.\r\n"
        "Send 'H' for help.\r\n"
    )

    def __init__(self):
        self.commands = {
            "!AC": commands.AccessControlCommand(),
            "!AS": commands.AlarmServerCommand(),
            "!GC": commands.GetConfigCommand(),
            "!GV": commands.SoftwareVersionCommand(),
            "!SA": commands.SetKap1Command(),
            "!SB": commands.SetKap2Command(),
            "!SC": commands.SetConfigCommand(),
            "!SD": commands.SetDeviceNameCommand(),
            "!SH": commands.SetLookupCommand(),
            "!SI": commands.SetIPCommand(),
            "!SK": commands.SetWatchdogCommand(),
            "!SN": commands.SetNameserverCommand(),
            "!SP": commands.SetPortsCommand(),
            "!SS": commands.SetSerialCommand(),
            "!RC": commands.RequestConnectCommand(),
            "!RR": commands.RequestRestartCommand(),
            "!WM": commands.WinkModuleCommand(),
        }

        self.help_command = commands.HelpCommand(self.commands)

    def respond(self, request):
        stripped_request = request.strip()

        if len(stripped_request) == 0:
            return ""  # idle

        split_request = stripped_request.split(" ", 1)
        command = split_request[0].upper()

        if len(command) > 3:
            return self.COMMAND_NOT_FOUND
        elif command.startswith("Q"):
            return  # quit

        params = None
        if len(split_request) > 1:
            params = split_request[1]

        if command.startswith("H"):
            return self.help_command.run(params)
        if command.startswith("!"):
            if command in list(self.commands.keys()):
                return self.commands[command].run(params)

        return self.COMMAND_NOT_FOUND



# Copyright (C) 2013  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import time

from gevent.server import StreamServer
import codecs
import socket
from struct import unpack
from conpot.protocols.s7comm.tpkt import TPKT
from conpot.protocols.s7comm.cotp import COTP as COTP_BASE_packet
from conpot.protocols.s7comm.cotp import COTP_ConnectionRequest
from conpot.protocols.s7comm.cotp import COTP_ConnectionConfirm
from conpot.protocols.s7comm.s7 import S7
import conpot.core as conpot_core
from conpot.core.protocol_wrapper import conpot_protocol
from lxml import etree

import logging
logger = logging.getLogger(__name__)


def cleanse_byte_string(packet):
    new_packet = packet.decode('latin-1').replace('b', '')
    return new_packet.encode('latin-1')

@conpot_protocol
class S7Server(object):
    def __init__(self, template, template_directory, args):

        self.timeout = 5
        self.ssl_lists = {}
        self.server = None
        S7.ssl_lists = self.ssl_lists
        self.start_time = None  # Initialize later
        dom = etree.parse(template)

        system_status_lists = dom.xpath('//s7comm/system_status_lists/*')
        for ssl in system_status_lists:
            ssl_id = ssl.attrib['id']
            ssl_dict = {}
            self.ssl_lists[ssl_id] = ssl_dict
            items = ssl.xpath('./*')
            for item in items:
                item_id = item.attrib['id']
                databus_key = item.xpath('./text()')[0] if len(item.xpath('./text()')) else ''
                ssl_dict[item_id] = databus_key

        logger.debug('Conpot debug info: S7 SSL/SZL: {0}'.format(self.ssl_lists))
        logger.info('Conpot S7Comm initialized')

    def handle(self, sock, address):
        sock.settimeout(self.timeout)
        session = conpot_core.get_session('s7comm', address[0], address[1])

        self.start_time = time.time()
        logger.info('New S7 connection from {0}:{1}. ({2})'.format(address[0], address[1], session.id))
        session.add_event({'type': 'NEW_CONNECTION'})

        try:
            while True:

                data = sock.recv(4, socket.MSG_WAITALL)
                if len(data) == 0:
                    session.add_event({'type': 'CONNECTION_LOST'})
                    break

                _, _, length = unpack('!BBH', data[:4])
                # check for length
                if length <= 4:
                    logger.info('S7 error: Invalid length')
                    session.add_event({'error': 'S7 error: Invalid length'})
                    break
                data += sock.recv(length - 4, socket.MSG_WAITALL)

                tpkt_packet = TPKT().parse(cleanse_byte_string(data))
                cotp_base_packet = COTP_BASE_packet().parse(tpkt_packet.payload)
                if cotp_base_packet.tpdu_type == 0xe0:

                    # connection request
                    cotp_cr_request = COTP_ConnectionRequest().dissect(cotp_base_packet.payload)
                    logger.info('Received COTP Connection Request: dst-ref:{0} src-ref:{1} dst-tsap:{2} src-tsap:{3} '
                                'tpdu-size:{4}. ({5})'.format(cotp_cr_request.dst_ref, cotp_cr_request.src_ref,
                                                              cotp_cr_request.dst_tsap, cotp_cr_request.src_tsap,
                                                              cotp_cr_request.tpdu_size, session.id))

                    # confirm connection response
                    cotp_cc_response = COTP_ConnectionConfirm(cotp_cr_request.src_ref, cotp_cr_request.dst_ref, 0,
                                                              cotp_cr_request.src_tsap, cotp_cr_request.dst_tsap,
                                                              0x0a).assemble()

                    # encapsulate and transmit
                    cotp_resp_base_packet = COTP_BASE_packet(0xd0, 0, cotp_cc_response).pack()
                    tpkt_resp_packet = TPKT(3, cotp_resp_base_packet).pack()
                    sock.send(tpkt_resp_packet)

                    session.add_event({'request': codecs.encode(data, 'hex'),
                                       'response': codecs.encode(tpkt_resp_packet, 'hex')})

                    data = sock.recv(1024)

                    # another round of parsing payloads
                    tpkt_packet = TPKT().parse(data)
                    cotp_base_packet = COTP_BASE_packet().parse(tpkt_packet.payload)

                    if cotp_base_packet.tpdu_type == 0xf0:
                        logger.info('Received known COTP TPDU: {0}. ({1})'.format(cotp_base_packet.tpdu_type,
                                                                                   session.id))

                        # will throw exception if the packet does not contain the S7 magic number (0x32)
                        S7_packet = S7().parse(cotp_base_packet.trailer)
                        logger.info(
                            'Received S7 packet: magic:%s pdu_type:%s reserved:%s req_id:%s param_len:%s '
                            'data_len:%s result_inf:%s session_id:%s',
                            S7_packet.magic, S7_packet.pdu_type,
                            S7_packet.reserved, S7_packet.request_id,
                            S7_packet.param_length, S7_packet.data_length,
                            S7_packet.result_info, session.id)

                        # request pdu
                        if S7_packet.pdu_type == 1:

                            # 0xf0 == Request for connect / pdu negotiate
                            if S7_packet.param == 0xf0:

                                # create S7 response packet
                                s7_resp_negotiate_packet = S7(3, 0, S7_packet.request_id, 0,
                                                              S7_packet.parameters).pack()
                                # wrap s7 the packet in cotp
                                cotp_resp_negotiate_packet = COTP_BASE_packet(0xf0, 0x80,
                                                                              s7_resp_negotiate_packet).pack()
                                # wrap the cotp packet
                                tpkt_resp_packet = TPKT(3, cotp_resp_negotiate_packet).pack()
                                sock.send(tpkt_resp_packet)

                                session.add_event({'request': codecs.encode(data, 'hex'),
                                                   'response': codecs.encode(tpkt_resp_packet, 'hex')})

                                # handshake done, give some more data.
                                data = sock.recv(1024)

                                while data:
                                    tpkt_packet = TPKT().parse(data)
                                    cotp_base_packet = COTP_BASE_packet().parse(tpkt_packet.payload)

                                    if cotp_base_packet.tpdu_type == 0xf0:
                                        S7_packet = S7().parse(cotp_base_packet.trailer)
                                        logger.info(
                                            'Received S7 packet: magic:%s pdu_type:%s reserved:%s '
                                            'req_id:%s param_len:%s data_len:%s result_inf:%s session_id:%s',
                                            S7_packet.magic, S7_packet.pdu_type,
                                            S7_packet.reserved, S7_packet.request_id,
                                            S7_packet.param_length, S7_packet.data_length,
                                            S7_packet.result_info, session.id)

                                        response_param, response_data = S7_packet.handle()
                                        s7_resp_ssl_packet = S7(7, 0, S7_packet.request_id, 0, response_param,
                                                                response_data).pack()
                                        cotp_resp_ssl_packet = COTP_BASE_packet(0xf0, 0x80, s7_resp_ssl_packet).pack()
                                        tpkt_resp_packet = TPKT(3, cotp_resp_ssl_packet).pack()
                                        sock.send(tpkt_resp_packet)

                                        session.add_event({'request': codecs.encode(data, 'hex'),
                                                           'response': codecs.encode(tpkt_resp_packet, 'hex')})

                                    data = sock.recv(1024)
                    else:
                        logger.info(
                            'Received unknown COTP TPDU after handshake: {0}'.format(cotp_base_packet.tpdu_type))
                        session.add_event({'error': 'Received unknown COTP TPDU after handshake: {0}'.format(cotp_base_packet.tpdu_type)})
                else:
                    logger.info('Received unknown COTP TPDU before handshake: {0}'.format(cotp_base_packet.tpdu_type))
                    session.add_event({'error': 'Received unknown COTP TPDU before handshake: {0}'.format(cotp_base_packet.tpdu_type)})

        except socket.timeout:
            session.add_event({'type': 'CONNECTION_LOST'})
            logger.debug('Socket timeout, remote: {0}. ({1})'.format(address[0], session.id))
        except socket.error:
            session.add_event({'type': 'CONNECTION_LOST'})
            logger.debug('Connection reset by peer, remote: {0}. ({1})'.format(address[0], session.id))
        except Exception as e:
            logger.exception('Exception caught {0}, remote: {1}. ({2})'.format(e, address[0], session.id))

    def start(self, host, port):
        connection = (host, port)
        self.server = StreamServer(connection, self.handle)
        logger.info('S7Comm server started on: {0}'.format(connection))
        self.server.serve_forever()

    def stop(self):
        self.server.stop()


if __name__ == '__main__':
    import os
    test_template = os.getcwd() + '/../../templates/default/s7comm/s7comm.xml'
    test = S7Server(test_template, None, None)
    try:
        test.start('127.0.0.1', 9999)
    except KeyboardInterrupt:
        test.stop()


# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import gevent
import os
import unittest
import conpot
from gevent import socket
import conpot.core as conpot_core
from conpot.protocols.kamstrup.management_protocol.kamstrup_management_server import KamstrupManagementServer
from conpot.tests.data.kamstrup_management_data import RESPONSES


def check_command_resp_help_message(packet_type, help_msg_command, packet_msg_command, kamstrup_management_server):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(('127.0.0.1', kamstrup_management_server.server.server_port))
    _ = s.recv(1024)  # receive the banner
    s.sendall(help_msg_command)  # test the help command
    help_data = s.recv(1024)
    help_response = (help_data == RESPONSES['H'][packet_type])
    s.sendall(packet_msg_command)
    pkt_data = s.recv(1024)
    packet_resp = (pkt_data == RESPONSES[packet_type])
    s.close()
    return help_response and packet_resp


class TestKamstrupManagementProtocol(unittest.TestCase):
    """
        All tests work in similar way. We send a get command check for a valid reply. We send in set command and
        expect things to change in the databus.
    """

    def setUp(self):

        # clean up before we start...
        conpot_core.get_sessionManager().purge_sessions()

        # get the conpot directory
        self.dir_name = os.path.dirname(conpot.__file__)
        self.kamstrup_management_server = KamstrupManagementServer(
            self.dir_name + '/templates/kamstrup_382/kamstrup_management/kamstrup_management.xml', None, None
        )
        self.server_greenlet = gevent.spawn(self.kamstrup_management_server.start, '127.0.0.1', 0)

        # initialize the databus
        self.databus = conpot_core.get_databus()
        self.databus.initialize(self.dir_name + '/templates/kamstrup_382/template.xml')
        gevent.sleep(1)

    def tearDown(self):
        self.kamstrup_management_server.stop()
        gevent.joinall([self.server_greenlet])
        # tidy up (again)...
        conpot_core.get_sessionManager().purge_sessions()

    def test_help_command(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.kamstrup_management_server.server.server_port))
        _ = s.recv(1024)  # receive the banner
        s.sendall(b'H\r\n')  # test the help command
        data = s.recv(1024)
        s.close()
        self.assertEqual(data, RESPONSES['H']['H'])

    def test_set_config_command(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.kamstrup_management_server.server.server_port))
        _ = s.recv(1024)  # receive the banner
        s.sendall(b'H !SC\r\n')  # test the help command
        data = s.recv(1024)
        s.close()
        self.assertEqual(data, RESPONSES['H']['!SC'])

    def test_access_control_command(self):
        self.assertTrue(check_command_resp_help_message('!AC', b'H !AC\r\n', b'!AC 0 2 192.168.1.211\r\n',
                                                        self.kamstrup_management_server))

    def test_alarm_server_command(self):
        self.assertTrue(check_command_resp_help_message('!AS', b'H !AS\r\n', b'!AC 192.168.1.4 4000\r\n',
                                                        self.kamstrup_management_server))

    def test_get_config_command(self):
        self.assertTrue(check_command_resp_help_message('!GC', b'H !GC\r\n', b'!GC\r\n',
                                                        self.kamstrup_management_server))

    def test_get_software_version_command(self):
        self.assertTrue(check_command_resp_help_message('!GV', b'H !GV\r\n', b'!GV\r\n',
                                                        self.kamstrup_management_server))

    def test_set_kap1_command(self):
        # TODO: verify that values in the databus have actually changed!
        self.assertTrue(check_command_resp_help_message('!SA', b'H !SA\r\n', b'!SA 192168001002 61000\r\n',
                                                        self.kamstrup_management_server))

    def test_set_kap2_command(self):
        # TODO: verify that values in the databus have actually changed!
        self.assertTrue(check_command_resp_help_message('!SB', b'H !SB\r\n', b'!SB 192.168.1.2 61000\r\n',
                                                        self.kamstrup_management_server))

    def test_set_device_name_command(self):
        # TODO: verify that values in the databus have actually changed!
        self.assertTrue(check_command_resp_help_message('!SD', b'H !SD\r\n', b'!SD\r\n',
                                                        self.kamstrup_management_server))

    def test_set_lookup_command(self):
        # TODO: verify that values in the databus have actually changed!
        self.assertTrue(check_command_resp_help_message('!SH', b'H !SH\r\n', b'!SH hosting.kamstrup_meter.dk\r\n',
                                                        self.kamstrup_management_server))

    def test_set_ip_command(self):
        # TODO: verify that values in the databus have actually changed!
        self.assertTrue(check_command_resp_help_message('!SI', b'H !SI\r\n', b'!SI 192168001200\r\n',
                                                        self.kamstrup_management_server))

    def test_set_watchdog_command(self):
        # TODO: verify that values in the databus have actually changed!
        self.assertTrue(check_command_resp_help_message('!SK', b'H !SK\r\n', b'!SK 3600 60 10\r\n',
                                                        self.kamstrup_management_server))

    def test_set_name_server_command(self):
        # TODO: verify that values in the databus have actually changed!
        self.assertTrue(check_command_resp_help_message('!SN',
                                                        b'H !SN\r\n',
                                                        b'!SN 192168001200 192168001201 000000000000\r\n',
                                                        self.kamstrup_management_server))

    def test_set_ports_command(self):
        # TODO: verify that values in the databus have actually changed!
        self.assertTrue(check_command_resp_help_message('!SP', b'H !SP\r\n', b'!SP 50 1025 1026 50100\r\n',
                                                        self.kamstrup_management_server))

    def test_set_serial_command(self):
        # TODO: verify that values in the databus have actually changed!
        self.assertTrue(check_command_resp_help_message('!SS', b'H !SS\r\n', b'!SS B 115200,8,E,1,L\r\n',
                                                        self.kamstrup_management_server))

    def test_request_connect_command(self):
        self.assertTrue(check_command_resp_help_message('!RC', b'H !RC\r\n', b'!RC A 195.215.168.45\r\n',
                                                        self.kamstrup_management_server))


if __name__ == '__main__':
    unittest.main()


# Copyright (C) 2013  Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import gevent.monkey
gevent.monkey.patch_all()

import unittest
import tempfile
import shutil
from collections import namedtuple

import gevent
import os
from pysnmp.proto import rfc1902
import conpot
import conpot.core as conpot_core
from conpot.tests.helpers import snmp_client
from conpot.protocols.snmp.snmp_server import SNMPServer


class TestSNMPServer(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        self.host = '127.0.0.1'
        databus = conpot_core.get_databus()
        # get the current directory
        self.dir_name = os.path.dirname(conpot.__file__)
        databus.initialize(self.dir_name + '/templates/default/template.xml')
        args = namedtuple('FakeArgs', 'mibpaths raw_mib')
        args.mibpaths = [self.tmp_dir]
        args.raw_mib = [self.tmp_dir]
        self.snmp_server = SNMPServer(self.dir_name + '/templates/default/snmp/snmp.xml', 'none', args)
        self.server_greenlet = gevent.spawn(self.snmp_server.start, self.host, 0)
        gevent.sleep(1)
        self.port = self.snmp_server.get_port()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_snmp_get(self):
        """
        Objective: Test if we can get data via snmp_get
        """
        client = snmp_client.SNMPClient(self.host, self.port)
        oid = ((1, 3, 6, 1, 2, 1, 1, 1, 0), None)
        client.get_command(oid, callback=self.mock_callback)
        self.assertEqual("Siemens, SIMATIC, S7-200", self.result)

    def test_snmp_set(self):
        """
        Objective: Test if we can set data via snmp_set
        """
        client = snmp_client.SNMPClient(self.host, self.port)
        # syslocation
        oid = ((1, 3, 6, 1, 2, 1, 1, 6, 0), rfc1902.OctetString('TESTVALUE'))
        client.set_command(oid, callback=self.mock_callback)
        databus = conpot_core.get_databus()
        self.assertEqual('TESTVALUE', databus.get_value('sysLocation')._value.decode())

    def mock_callback(self, sendRequestHandle, errorIndication, errorStatus, errorIndex, varBindTable, cbCtx):
        self.result = None
        if errorIndication:
            self.result = errorIndication
        elif errorStatus:
            self.result = errorStatus.prettyPrint()
        else:
            for oid, val in varBindTable:
                self.result = val.prettyPrint()


if __name__ == '__main__':
    unittest.main()


# Copyright (C) 2014  Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import json
import logging
import socket

import requests
from requests.exceptions import Timeout, ConnectionError


logger = logging.getLogger(__name__)


def _verify_address(addr):
    try:
        socket.inet_aton(addr)
        return True
    except (socket.error, UnicodeEncodeError, TypeError):
        return False


def _fetch_data(urls):
    # we only want warning+ messages from the requests module
    logging.getLogger("requests").setLevel(logging.WARNING)
    for url in urls:
        try:
            req = requests.get(url, timeout=5)
            if req.status_code == 200:
                data = req.text.strip()
                if data is None or not _verify_address(data):
                    continue
                else:
                    return data
            else:
                raise ConnectionError
        except (Timeout, ConnectionError):
            logger.warning('Could not fetch public ip from %s', url)
    return None


def get_ext_ip(config=None, urls=None):
    if config:
        urls = json.loads(config.get('fetch_public_ip', 'urls'))
    public_ip = _fetch_data(urls)
    if public_ip:
        logger.info('Fetched %s as external ip.', public_ip)
    else:
        logger.warning('Could not fetch public ip: %s', public_ip)
    return public_ip


if __name__ == "__main__":
    print((get_ext_ip(urls=["https://api.ipify.org", "http://127.0.0.1:8000"])))



# Copyright (C) 2014 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from .session_manager import SessionManager
from .virtual_fs import VirtualFS
from typing import Tuple, Union, Optional
from .internal_interface import Interface

sessionManager = SessionManager()
virtualFS = VirtualFS()
core_interface = Interface()

# databus related  --


def get_sessionManager():
    return sessionManager


def get_databus():
    return sessionManager._databus


def get_session(*args, **kwargs):
    return sessionManager.get_session(*args, **kwargs)

# file-system related  --


def initialize_vfs(fs_path=None, data_fs_path=None, temp_dir=None):
    return virtualFS.initialize_vfs(fs_path=fs_path, data_fs_path=data_fs_path, temp_dir=temp_dir)


def add_protocol(protocol_name: str,
                 data_fs_subdir: str,
                 vfs_dst_path: str,
                 src_path=None,
                 owner_uid: Optional[int] = 0,
                 group_gid: Optional[int] = 0,
                 perms: Optional[oct] = 0o755) -> Tuple:
    return virtualFS.add_protocol(protocol_name, data_fs_subdir, vfs_dst_path, src_path, owner_uid, group_gid, perms)


def get_vfs(protocol_name: Optional[str] = None) -> Union[virtual_fs.AbstractFS, Tuple]:
    """
    Get the File System.
    :param protocol_name: Name of the protocol to be fetched
    """
    if protocol_name:
        return virtualFS._conpot_vfs[protocol_name]
    else:
        return virtualFS.protocol_fs


def close_fs():
    """Close the file system. Remove all the temp files."""
    virtualFS.close()

# internal-interface related   --


def get_interface():
    return globals()['core_interface']



"""
Some python3 fixtures - helper methods for handy conversions + fix ssl
"""
from datetime import datetime
from slugify import slugify


def sanitize_file_name(name, host, port):
    """
    Ensure that file_name is legal. Slug the filename and store it onto the server.
    This would ensure that there are no duplicates as far as writing a file is concerned. Also client addresses are
    noted so that one can verify which client uploaded the file.
    :param name: Name of the file
    :param host: host/client address
    :param port port/client port
    :type name: str
    """
    return '(' + host + ', ' + str(port) + ')-' + datetime.now().strftime("%Y-%m-%d %H:%M:%S") + '-' + slugify(name)


# py3 chr
def chr_py3(x): return bytearray((x,))


# covert a number to an ascii byte string
def number_to_bytes(x): return x if isinstance(x, bytes) else bytes(str(int(x)), encoding='ascii')


# convert a string to an ascii byte string
def str_to_bytes(x): return x if isinstance(x, bytes) else str(x).encode('ascii')


# pack a short int
def pack_short_int(x): return x if isinstance(x, bytes) else x.to_bytes(2, byteorder='big')


# unpack a short int
def unpack_short_int(x): return int.from_bytes(x,  byteorder='big')


months_map = {1: 'Jan', 2: 'Feb', 3: 'Mar', 4: 'Apr', 5: 'May', 6: 'Jun', 7: 'Jul', 8: 'Aug', 9: 'Sep', 10: 'Oct',
              11: 'Nov', 12: 'Dec'}


# https://www.bountysource.com/issues/4335201-ssl-broken-for-python-2-7-9
# Kudos to Eugene for this workaround!
def fix_sslwrap():
    # Re-add sslwrap to Python 2.7.9
    import inspect
    __ssl__ = __import__('ssl')

    try:
        _ssl = __ssl__._ssl
    except AttributeError:
        _ssl = __ssl__._ssl2

    def new_sslwrap(sock, server_side=False, keyfile=None, certfile=None, cert_reqs=__ssl__.CERT_NONE, ssl_version=__ssl__.PROTOCOL_SSLv23, ca_certs=None, ciphers=None):
        context = __ssl__.SSLContext(ssl_version)
        context.verify_mode = cert_reqs or __ssl__.CERT_NONE
        if ca_certs:
            context.load_verify_locations(ca_certs)
        if certfile:
            context.load_cert_chain(certfile, keyfile)
        if ciphers:
            context.set_ciphers(ciphers)

        caller_self = inspect.currentframe().f_back.f_locals['self']
        return context._wrap_socket(sock, server_side=server_side, ssl_sock=caller_self)

    if not hasattr(_ssl, 'sslwrap'):
        _ssl.sslwrap = new_sslwrap


# Copyright (C) 2017  Patrick Reichenberger (University of Passau) <patrick.reichenberger@t-online.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


from conpot.protocols.IEC104.errors import *


# direction either "m" for Monitor or "c" for control. Empty if irrelevant
# identical for some process information in monitor direction
def check_information_without_time(frame, direction):
    type_id = int(frame.getfieldval("TypeID"), 16)
    if frame.getfieldval("COT") in (2, 3, 5, 11, 12, 20):
        pass
    else:
        raise InvalidFieldValueException("Invalid COT for ASDU type " + str(type_id) + ".")


# identical for some process information in monitor direction
def check_information_with_time(frame, direction):
    type_id = int(frame.getfieldval("TypeID"), 16)
    if frame.getfieldval("SQ") == 0:
        if frame.getfieldval("COT") in (3, 5, 11, 12):
            pass
        else:
            raise InvalidFieldValueException("Invalid COT for ASDU type " + str(type_id) + ".")
    else:
        raise InvalidFieldValueException("SQ=1 not supported for ASDU type " + str(type_id) + ".")


def check_asdu_1(frame, direction):
    check_information_without_time(frame, direction)


def check_asdu_2(frame, direction):
    check_information_with_time(frame, direction)


def check_asdu_3(frame, direction):
    check_information_without_time(frame, direction)


def check_asdu_4(frame, direction):
    check_information_with_time(frame, direction)


def check_asdu_11(frame, direction):
    check_information_without_time(frame, direction)


def check_asdu_12(frame, direction):
    type_id = int(frame.getfieldval("TypeID"), 16)
    if frame.getfieldval("SQ") == 0:
        if frame.getfieldval("COT") in (3, 5):
            pass
        else:
            raise InvalidFieldValueException("Invalid COT for ASDU type " + str(type_id) + ".")
    else:
        raise InvalidFieldValueException("SQ=1 not supported for ASDU type " + str(type_id) + ".")


def check_asdu_13(frame, direction):
    check_information_without_time(frame, direction)


def check_asdu_14(frame, direction):
    type_id = int(frame.getfieldval("TypeID"), 16)
    if frame.getfieldval("SQ") == 0:
        if frame.getfieldval("COT") in (2, 3, 5, 11, 12, 20):
            pass
        else:
            raise InvalidFieldValueException("Invalid COT for ASDU type " + str(type_id) + ".")
    else:
        raise InvalidFieldValueException("SQ=1 not supported for ASDU type " + str(type_id) + ".")


def check_asdu_30(frame, direction):
    check_information_with_time(frame, direction)


def check_asdu_31(frame, direction):
    check_information_with_time(frame, direction)


def check_asdu_35(frame, direction):
    type_id = int(frame.getfieldval("TypeID"), 16)
    if frame.getfieldval("SQ") == 0:
        if frame.getfieldval("COT") in (3, 5):
            pass
        else:
            raise InvalidFieldValueException("Invalid COT for ASDU type " + str(type_id) + ".")
    else:
        raise InvalidFieldValueException("SQ=1 not supported for ASDU type " + str(type_id) + ".")


def check_asdu_36(frame, direction):
    type_id = int(frame.getfieldval("TypeID"), 16)
    if frame.getfieldval("SQ") == 0:
        if frame.getfieldval("COT") in (2, 3, 5, 11, 12, 20):
            pass
        else:
            raise InvalidFieldValueException("Invalid COT for ASDU type " + str(type_id) + ".")
    else:
        raise InvalidFieldValueException("SQ=1 not supported for ASDU type " + str(type_id) + ".")


# identical for process information in control direction
def check_command(frame, direction):
    type_id = int(frame.getfieldval("TypeID"))
    if frame.getfieldval("SQ") != 0:
        raise InvalidFieldValueException("SQ=1 not supported for ASDU type " + str(type_id) + ".")
    if frame.getfieldval("COT") in (6, 8) and direction == "c":
        pass
    elif frame.getfieldval("COT") in (7, 9, 10, 44, 45, 46, 47) and direction == "m":
        pass
    else:
        raise InvalidFieldValueException("Invalid COT for ASDU type " + str(type_id) + ".")
    number_of_objects = frame.getfieldval("NoO")
    if number_of_objects != 1:
        raise InvalidFieldValueException("Only one object allowed for ASDU type " + str(type_id) + ".")


def check_asdu_45(frame, direction):
    check_command(frame, direction)
    if direction == "c" and frame.getfieldval("LenAPDU") != 14:
        raise InvalidFieldValueException("Illogical length")


def check_asdu_46(frame, direction):
    check_command(frame, direction)


def check_asdu_47(frame, direction):
    check_command(frame, direction)


def check_asdu_48(frame, direction):
    check_command(frame, direction)


def check_asdu_49(frame, direction):
    check_command(frame, direction)


def check_asdu_50(frame, direction):
    check_command(frame, direction)


def check_asdu_51(frame, direction):
    check_command(frame, direction)


def check_asdu_100(frame, direction):
    type_id = int(frame.getfieldval("TypeID"))
    number_of_objects = frame.getfieldval("NoO")
    cause_of_transmission = frame.getfieldval("COT")
    qualif_of_inro = frame.getfieldval("QOI")
    if frame.getfieldval("SQ") != 0:
        raise InvalidFieldValueException("SQ=1 not supported for ASDU type " + str(type_id) + ".")
    if number_of_objects != 1:
        raise InvalidFieldValueException("Only one object allowed for ASDU type " + str(type_id) + ".")
    if cause_of_transmission in (6, 8) and direction == "c":
        pass
    elif cause_of_transmission in (7, 9, 10, 44, 45, 46, 47) and direction == "m":
        pass
    else:
        raise InvalidFieldValueException("Invalid COT for ASDU type " + str(type_id) + ".")
    if frame.getfieldval("IOA") != 0:
        raise InvalidFieldValueException("IOA not 0 for ASDU type " + str(type_id) + ".")
    if qualif_of_inro not in (0, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36):
        raise InvalidFieldValueException("Invalid QOI " + str(qualif_of_inro) + " for ASDU type " + str(type_id) + ".")



# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from enum import Enum

REQUEST_MAGIC = 0x80
RESPONSE_MAGIC = 0x40
EOT_MAGIC = 0x0d
ESCAPE = 0x1b

# what is 0x06?
NEED_ESCAPE = [0x06, EOT_MAGIC, ESCAPE, RESPONSE_MAGIC, REQUEST_MAGIC]

UNITS = {
    0: 'None',
    1: 'Wh',
    2: 'kWh',
    3: 'MWh',
    4: 'GWh',
    5: 'j',
    6: 'kj',
    7: 'Mj',
    8: 'Gj',
    9: 'Cal',
    10: 'kCal',
    11: 'MCal',
    12: 'GCal',
    13: 'varh',
    14: 'kvarh',
    15: 'Mvarh',
    16: 'Gvarh',
    17: 'VAh',
    18: 'kVAh',
    19: 'MVAh',
    20: 'GVAh',
    21: 'W',
    22: 'kW',
    23: 'MW',
    24: 'GW',
    25: 'var',
    26: 'kvar',
    27: 'MVar',
    28: 'Gvar',
    29: 'VA',
    30: 'kVA',
    31: 'MVA',
    32: 'GVA',
    33: 'V',
    34: 'A',
    35: 'kV',
    36: 'kA',
    37: 'C',
    38: 'K',
    39: 'I',
    40: 'm3',
    41: 'I_h',
    42: 'm3_h',
    43: 'm3xC',
    44: 'ton',
    45: 'ton_h',
    46: 'h',
    47: 'clock',  # hh:mm:ss
    48: 'date1',  # yy:mm:dd
    49: 'date2',  # yyyy:mm:dd
    50: 'date3',  # mm:dd
    51: 'number',
    52: 'bar',
    53: 'RTC',
    54: 'ASCII',
    55: 'm3x10',
    56: 'tonx10',
    57: 'GJx10',
    58: 'minutes',
    59: 'Bitfield',
    60: 's',
    61: 'ms',
    62: 'days',
    63: 'RTC_Q',
    64: 'Datetime',
    65: 'imp_L',
    66: 'L_imp',
    67: 'Hz',
    68: 'Degree',
    69: 'Percent',
    70: 'USgal',
    71: 'USgal_min',
    72: 'KamDateTime',
    73: 'IPv4Address',
    74: 'IPv6Address'
}


class MeterTypes(Enum):
    Unknown = 0,
    K382M = 1,
    K162M = 2,
    K351C = 3,
    OMNIA = 4,
    # where does 382J fit in? together with 382M?



# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import unittest
import os
import gevent
from gevent.server import StreamServer
from gevent.socket import socket
from gevent.ssl import wrap_socket
from conpot.helpers import fix_sslwrap
import conpot
from conpot.emulators.proxy import Proxy
from conpot.protocols.misc.ascii_decoder import AsciiDecoder
# gevent.monkey.patch_all()

package_directory = os.path.dirname(os.path.abspath(conpot.__file__))


class TestProxy(unittest.TestCase):
    def test_proxy(self):
        self.test_input = 'Hiya, this is a test'.encode('utf-8')
        mock_service = StreamServer(('127.0.0.1', 0), self.echo_server)
        gevent.spawn(mock_service.start)
        gevent.sleep(1)

        proxy = Proxy('proxy', '127.0.0.1', mock_service.server_port)
        server = proxy.get_server('127.0.0.1', 0)
        gevent.spawn(server.start)
        gevent.sleep(1)

        s = socket()
        s.connect(('127.0.0.1', server.server_port))
        s.sendall(self.test_input)
        received = s.recv(len(self.test_input))
        self.assertEqual(self.test_input, received)
        mock_service.stop(1)

    def test_ssl_proxy(self):
        fix_sslwrap()
        self.test_input = 'Hiya, this is a test'.encode('utf-8')
        keyfile = os.path.join(package_directory, 'templates/default/ssl/ssl.key')
        certfile = os.path.join(package_directory, 'templates/default/ssl/ssl.crt')

        mock_service = StreamServer(('127.0.0.1', 0), self.echo_server, keyfile=keyfile, certfile=certfile)
        gevent.spawn(mock_service.start)
        gevent.sleep(1)

        proxy = Proxy('proxy', '127.0.0.1', mock_service.server_port, keyfile=keyfile, certfile=certfile)
        server = proxy.get_server('127.0.0.1', 0)
        gevent.spawn(server.start)
        gevent.sleep(1)

        s = wrap_socket(socket(), keyfile, certfile)
        s.connect(('127.0.0.1', server.server_port))
        s.sendall(self.test_input)
        received = s.recv(len(self.test_input))
        self.assertEqual(self.test_input, received)
        mock_service.stop(1)

    def test_ascii_decoder(self):
        test_decoder = AsciiDecoder()
        # should not raise a UnicodeDecodeError
        self.assertTrue((test_decoder.decode_in(b'\x80abc') == b'\xef\xbf\xbdabc') and
                        (test_decoder.decode_out(b'\x80abc') == b'\xef\xbf\xbdabc'))

    def test_proxy_with_decoder(self):
        self.test_input = 'Hiya, this is a test'.encode('utf-8')
        mock_service = StreamServer(('127.0.0.1', 0), self.echo_server)
        gevent.spawn(mock_service.start)
        gevent.sleep(1)

        proxy = Proxy('proxy', '127.0.0.1', mock_service.server_port,
                      decoder='conpot.protocols.misc.ascii_decoder.AsciiDecoder')
        server = proxy.get_server('127.0.0.1', 0)
        gevent.spawn(server.start)
        gevent.sleep(1)

        s = socket()
        s.connect(('127.0.0.1', server.server_port))
        s.sendall(self.test_input)
        received = s.recv(len(self.test_input))
        self.assertEqual(self.test_input, received)
        mock_service.stop(1)

    def test_ssl_proxy_with_decoder(self):
        fix_sslwrap()
        self.test_input = 'Hiya, this is a test'.encode('utf-8')
        keyfile = os.path.join(package_directory, 'templates/default/ssl/ssl.key')
        certfile = os.path.join(package_directory, 'templates/default/ssl/ssl.crt')

        mock_service = StreamServer(('127.0.0.1', 0), self.echo_server, keyfile=keyfile, certfile=certfile)
        gevent.spawn(mock_service.start)
        gevent.sleep(1)

        proxy = Proxy('proxy', '127.0.0.1', mock_service.server_port,
                      decoder='conpot.protocols.misc.ascii_decoder.AsciiDecoder', keyfile=keyfile, certfile=certfile)
        server = proxy.get_server('127.0.0.1', 0)
        gevent.spawn(server.start)
        gevent.sleep(1)

        s = wrap_socket(socket(), keyfile, certfile)
        s.connect(('127.0.0.1', server.server_port))
        s.sendall(self.test_input)
        received = s.recv(len(self.test_input))
        self.assertEqual(self.test_input, received)
        mock_service.stop(1)

    def echo_server(self, sock, address):
        r = sock.recv(len(self.test_input))
        sock.send(r)


if __name__ == '__main__':
    unittest.main()


# Copyright (C) 2015  Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Service support based on gaspot.py [https://github.com/sjhilt/GasPot]
Original authors: Kyle Wilhoit and Stephen Hilt
"""

from gevent.server import StreamServer
import datetime
import random
import conpot
import conpot.core as conpot_core
from conpot.core.protocol_wrapper import conpot_protocol
from conpot.helpers import str_to_bytes

import logging
logger = logging.getLogger(__name__)

# 9999 indicates that the command was not understood and
# FF1B is the checksum for the 9999
AST_ERROR = "9999FF1B\n"


@conpot_protocol
class GuardianASTServer(object):
    def __init__(self, template, template_directory, args):
        self.server = None
        self.databus = conpot_core.get_databus()
        # dom = etree.parse(template)
        self.fill_offset_time = datetime.datetime.utcnow()
        logger.info('Conpot GuardianAST initialized')

    def handle(self, sock, addr):
        session = conpot_core.get_session('guardian_ast', addr[0], addr[1])
        logger.info('New GuardianAST connection from %s:%d. (%s)', addr[0], addr[1], session.id)
        session.add_event({'type': 'NEW_CONNECTION'})
        current_time = datetime.datetime.utcnow()
        fill_start = self.fill_offset_time - datetime.timedelta(minutes=313)
        fill_stop = self.fill_offset_time - datetime.timedelta(minutes=303)
        # Default Product names, change based off country needs
        product1 = self.databus.get_value('product1').ljust(22)
        product1 = self.databus.get_value('product1').ljust(22)
        product2 = self.databus.get_value('product2').ljust(22)
        product3 = self.databus.get_value('product3').ljust(22)
        product4 = self.databus.get_value('product4').ljust(22)

        # Create random Numbers for the volumes
        #
        # this will crate an initial Volume and then the second value based
        # off the orig value.
        vol1 = self.databus.get_value('vol1')
        vol1tc = random.randint(vol1, vol1+200)
        vol2 = self.databus.get_value('vol2')
        vol2tc = random.randint(vol2, vol2+200)
        vol3 = self.databus.get_value('vol3')
        vol3tc = random.randint(vol3, vol3+200)
        vol4 = self.databus.get_value('vol4')
        vol4tc = random.randint(vol4, vol4+200)

        # unfilled space ULLAGE
        ullage1 = str(self.databus.get_value('ullage1'))
        ullage2 = str(self.databus.get_value('ullage2'))
        ullage3 = str(self.databus.get_value('ullage3'))
        ullage4 = str(self.databus.get_value('ullage3'))

        # Height of tank
        height1 = str(self.databus.get_value('height1')).ljust(5, '0')
        height2 = str(self.databus.get_value('height2')).ljust(5, '0')
        height3 = str(self.databus.get_value('height3')).ljust(5, '0')
        height4 = str(self.databus.get_value('height4')).ljust(5, '0')

        # Water in tank, this is a variable that needs to be low
        h2o1 = str(self.databus.get_value('h2o1')).ljust(4, '0')
        h2o2 = str(self.databus.get_value('h2o2')).ljust(4, '0')
        h2o3 = str(self.databus.get_value('h2o3')).ljust(4, '0')
        h2o4 = str(self.databus.get_value('h2o4')).ljust(4, '0')

        # Temperature of the tank, this will need to be between 50 - 60
        temp1 = str(self.databus.get_value('temp1')).ljust(5, '0')
        temp2 = str(self.databus.get_value('temp2')).ljust(5, '0')
        temp3 = str(self.databus.get_value('temp3')).ljust(5, '0')
        temp4 = str(self.databus.get_value('temp4')).ljust(5, '0')

        station = self.databus.get_value('station_name')

        # This function is to set-up up the message to be sent upon a successful I20100 command being sent
        # The final message is sent with a current date/time stamp inside of the main loop.
        def I20100():
            ret = '\nI20100\n' + str(current_time.strftime('%m/%d/%Y %H:%M'))
            ret += '\n\n' + station + '\n\n\n\nIN-TANK INVENTORY\n\n'
            ret += 'TANK PRODUCT             VOLUME TC VOLUME   ULLAGE   HEIGHT    WATER     TEMP'
            ret += '\n  1  ' + product1 + str(vol1) + '      ' + str(vol1tc) + '     ' + ullage1 + '    ' + height1 + '     ' + h2o1 + '    ' + temp1
            ret += '\n  2  ' + product2 + str(vol2) + '      ' + str(vol2tc) + '     ' + ullage2 + '    ' + height2 + '     ' + h2o2 + '    ' + temp2
            ret += '\n  3  ' + product3 + str(vol3) + '      ' + str(vol3tc) + '     ' + ullage3 + '    ' + height3 + '     ' + h2o3 + '    ' + temp3
            ret += '\n  4  ' + product4 + str(vol4) + '      ' + str(vol4tc) + '     ' + ullage4 + '    ' + height4 + '     ' + h2o4 + '    ' + temp4
            ret += '\n'
            return ret

        ###########################################################################
        #
        # Only one Tank is listed currently in the I20200 command
        #
        ###########################################################################
        def I20200():
            ret = '\nI20200\n' + str(current_time.strftime('%m/%d/%Y %H:%M'))
            ret += '\n\n' + station + '\n\n\n\nDELIVERY REPORT\n\n'
            ret += 'T 1:' + product1 + '\nINCREASE   DATE / TIME             GALLONS TC GALLONS WATER  TEMP DEG F  HEIGHT\n\n'

            ret += '      END: ' + str(fill_stop.strftime('%m/%d/%Y %H:%M')) + '         ' + str(vol1 + 300) + '       ' + str(vol1tc + 300) + '   ' + h2o1 + '      ' + temp1 + '    ' + height1 + '\n'
            ret += '    START: ' + str(fill_start.strftime('%m/%d/%Y %H:%M')) + '         ' + str(vol1 - 300) + '       ' + str(vol1tc - 300) + '   ' + h2o1 + '      ' + temp1 + '    ' + str(float(height1) - 23) + '\n'
            ret += '   AMOUNT:                          ' + str(vol1) + '       ' + str(vol1tc) + '\n\n'
            return ret

        ###########################################################################
        #
        # I20300 In-Tank Leak Detect Report
        #
        ###########################################################################
        def I20300():
            ret = '\nI20300\n' + str(current_time.strftime('%m/%d/%Y %H:%M'))
            ret += '\n\n' + station + '\n\n\n'
            ret += 'TANK 1    ' + product1 + '\n    TEST STATUS: OFF\nLEAK DATA NOT AVAILABLE ON THIS TANK\n\n'
            ret += 'TANK 2    ' + product2 + '\n    TEST STATUS: OFF\nLEAK DATA NOT AVAILABLE ON THIS TANK\n\n'
            ret += 'TANK 3    ' + product3 + '\n    TEST STATUS: OFF\nLEAK DATA NOT AVAILABLE ON THIS TANK\n\n'
            ret += 'TANK 4    ' + product4 + '\n    TEST STATUS: OFF\nLEAK DATA NOT AVAILABLE ON THIS TANK\n\n'
            return ret

        ###########################################################################
        # Shift report command I20400 only one item in report at this time,
        # but can always add more if needed
        ###########################################################################
        def I20400():
            ret = '\nI20400\n' + str(current_time.strftime('%m/%d/%Y %H:%M'))
            ret += '\n\n' + station + '\n\n\n\nSHIFT REPORT\n\n'
            ret += 'SHIFT 1 TIME: 12:00 AM\n\nTANK PRODUCT\n\n'
            ret += '  1  ' + product1 + ' VOLUME TC VOLUME  ULLAGE  HEIGHT  WATER   TEMP\n'
            ret += 'SHIFT  1 STARTING VALUES      ' + str(vol1) + '     ' + str(vol1tc) + '    ' + ullage1 + '   ' + height1 + '   ' + h2o1 + '    ' + temp1 + '\n'
            ret += '         ENDING VALUES        ' + str(vol1 + 940) + '     ' + str(vol1tc + 886) + '    ' + str(int(ullage1) + 345) + '   ' + str(float(height1) + 53) + '  ' + h2o1 + '    ' + temp1 + '\n'
            ret += '         DELIVERY VALUE          0\n'
            ret += '         TOTALS                940\n\n'
            return ret

        ###########################################################################
        # I20500 In-Tank Status Report
        ###########################################################################
        def I20500():
            ret = '\nI20500\n' + str(current_time.strftime('%m/%d/%Y %H:%M'))
            ret += '\n\n\n' + station + '\n\n\n'
            ret += 'TANK   PRODUCT                 STATUS\n\n'
            ret += '  1    ' + product1 + '  NORMAL\n\n'
            ret += '  2    ' + product2 + '  HIGH WATER ALARM\n'
            ret += '                               HIGH WATER WARNING\n\n'
            ret += '  3    ' + product3 + '  NORMAL\n\n'
            ret += '  4    ' + product4 + '  NORMAL\n\n'
            return ret

        while True:
            try:
                # Get the initial data
                request = sock.recv(4096)
                # The connection has been closed
                if not request:
                    break
                while not (b'\n' in request or b'00' in request):
                    request += sock.recv(4096)
                # if first value is not ^A then do nothing
                # thanks John(achillean) for the help
                if request[:1] != b'\x01':
                    logger.info('Non ^A command attempt %s:%d. (%s)', addr[0], addr[1], session.id)
                    break
                # if request is less than 6, than do nothing
                if len(request) < 6:
                    logger.info('Invalid command attempt %s:%d. (%s)', addr[0], addr[1], session.id)
                    break

                cmds = {"I20100": I20100, "I20200": I20200, "I20300": I20300, "I20400": I20400, "I20500": I20500}
                cmd = request[1:7].decode()  # strip ^A and \n out
                response = None
                if cmd in cmds:
                    logger.info('%s command attempt %s:%d. (%s)', cmd, addr[0], addr[1], session.id)
                    response = cmds[cmd]()
                elif cmd.startswith("S6020"):
                    # change the tank name
                    if cmd.startswith("S60201"):
                        # split string into two, the command, and the data
                        TEMP = request.split(b'S60201')
                        # if length is less than two, print error
                        if len(TEMP) < 2:
                            response = AST_ERROR
                        # Else the command was entered correctly and continue
                        else:
                            # Strip off the carrage returns and new lines
                            TEMP1 = TEMP[1].rstrip(b'\r\n').decode()
                            # if Length is less than 22
                            if len(TEMP1) < 22:
                                # pad the result to have 22 chars
                                product1 = TEMP1.ljust(22)
                            elif len(TEMP1) > 22:
                                # else only print 22 chars if the result was longer
                                product1 = TEMP1[:20] + "  "
                            else:
                                # else it fits fine (22 chars)
                                product1 = TEMP1
                        logger.info('S60201: %s command attempt %s:%d. (%s)', TEMP1, addr[0], addr[1], session.id)
                    # Follows format for S60201 for comments
                    elif cmd.startswith("S60202"):
                        TEMP = request.split(b'S60202')
                        if len(TEMP) < 2:
                            response = AST_ERROR
                        else:
                            TEMP1 = TEMP[1].rstrip(b'\r\n').decode()
                            if len(TEMP1) < 22:
                                product2 = TEMP1.ljust(22)
                            elif len(TEMP1) > 22:
                                product2 = TEMP1[:20] + "  "
                            else:
                                product2 = TEMP1
                        logger.info('S60202: %s command attempt %s:%d. (%s)', TEMP1, addr[0], addr[1], session.id)
                    # Follows format for S60201 for comments
                    elif cmd.startswith("S60203"):
                        TEMP = request.split(b'S60203')
                        if len(TEMP) < 2:
                            response = AST_ERROR
                        else:
                            TEMP1 = TEMP[1].rstrip(b'\r\n').decode()
                            if len(TEMP1) < 22:
                                product3 = TEMP1.ljust(22)
                            elif len(TEMP1) > 22:
                                product3 = TEMP1[:20] + "  "
                            else:
                                product3 = TEMP1
                        logger.info('S60203: %s command attempt %s:%d. (%s)', TEMP1, addr[0], addr[1], session.id)
                    # Follows format for S60201 for comments
                    elif cmd.startswith("S60204"):
                        TEMP = request.split(b'S60204')
                        if len(TEMP) < 2:
                            response = AST_ERROR
                        else:
                            TEMP1 = TEMP[1].rstrip(b'\r\n').decode()
                            if len(TEMP1) < 22:
                                product4 = TEMP1.ljust(22)
                            elif len(TEMP1) > 22:
                                product4 = TEMP1[:20] + "  "
                            else:
                                product4 = TEMP1
                        logger.info('S60204: %s command attempt %s:%d. (%s)', TEMP1, addr[0], addr[1], session.id)
                    # Follows format for S60201 for comments
                    elif cmd.startswith("S60200"):
                        TEMP = request.split(b'S60200')
                        if len(TEMP) < 2:
                            response = AST_ERROR
                        else:
                            TEMP1 = TEMP[1].rstrip(b'\r\n').decode()
                            if len(TEMP1) < 22:
                                product1 = TEMP1.ljust(22)
                                product2 = TEMP1.ljust(22)
                                product3 = TEMP1.ljust(22)
                                product4 = TEMP1.ljust(22)
                            elif len(TEMP1) > 22:
                                product1 = TEMP1[:20] + "  "
                                product2 = TEMP1[:20] + "  "
                                product3 = TEMP1[:20] + "  "
                                product4 = TEMP1[:20] + "  "
                            else:
                                product1 = TEMP1
                                product2 = TEMP1
                                product3 = TEMP1
                                product4 = TEMP1
                        logger.info('S60200: %s command attempt %s:%d. (%s)', TEMP1, addr[0], addr[1], session.id)
                    else:
                        response = AST_ERROR
                else:
                    response = AST_ERROR
                    # log what was entered
                    logger.info('%s command attempt %s:%d. (%s)', request, addr[0], addr[1], session.id)
                if response:
                    sock.send(str_to_bytes(response))
                session.add_event({"type": "AST {0}".format(cmd), "request": request, "response": response})
            except Exception as e:
                logger.exception(('Unknown Error: {}'.format(str(e))))
        logger.info('GuardianAST client disconnected %s:%d. (%s)', addr[0], addr[1], session.id)
        session.add_event({'type': 'CONNECTION_LOST'})

    def start(self, host, port):
        connection = (host, port)
        self.server = StreamServer(connection, self.handle)
        logger.info('GuardianAST server started on: {0}'.format(connection))
        self.server.serve_forever()

    def stop(self):
        self.server.stop()


if __name__ == '__main__':
    # Set vars for connection information
    TCP_IP = '127.0.0.1'
    TCP_PORT = 10001
    import os
    dir_name = os.path.dirname(conpot.__file__)
    server = GuardianASTServer(None, None, None)
    server.databus.initialize(dir_name + '/templates/guardian_ast/template.xml')
    try:
        server.start(TCP_IP, TCP_PORT)
    except KeyboardInterrupt:
        server.stop()



class ParseException(Exception):

    def __init__(self, protocol, reason, payload=''):
        self.proto = protocol
        self.reason = reason
        self.payload = payload

    def __str__(self):
        return "DissectException: proto:{0} reason:{1}".format(self.proto, self.reason)


class AssembleException(Exception):

    def __init__(self, protocol, reason, payload=''):
        self.proto = protocol
        self.reason = reason
        self.payload = payload

    def __str__(self):
        return "AssembleException: proto:{0} reason:{1}".format(self.proto, self.reason)



# Copyright (C) 2015  Peter Sooky <xsooky00@stud.fit.vubtr.cz>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Author: Peter Sooky <xsooky00@stud.fit.vubtr.cz>
# Brno University of Technology, Faculty of Information Technology

import gevent
from gevent import monkey; gevent.monkey.patch_all()

import socket
from lxml import etree
from gevent.server import DatagramServer
from bacpypes.local.device import LocalDeviceObject
from bacpypes.apdu import APDU
from bacpypes.pdu import PDU
from bacpypes.errors import DecodingError
import conpot.core as conpot_core
from conpot.protocols.bacnet.bacnet_app import BACnetApp
from conpot.core.protocol_wrapper import conpot_protocol
import logging
logger = logging.getLogger(__name__)


@conpot_protocol
class BacnetServer(object):
    def __init__(self, template, template_directory, args):
        self.dom = etree.parse(template)
        device_info_root = self.dom.xpath('//bacnet/device_info')[0]
        name_key = device_info_root.xpath('./device_name/text()')[0]
        id_key = device_info_root.xpath('./device_identifier/text()')[0]
        vendor_name_key = device_info_root.xpath('./vendor_name/text()')[0]
        vendor_identifier_key = device_info_root.xpath('./vendor_identifier/text()')[0]
        apdu_length_key = device_info_root.xpath('./max_apdu_length_accepted/text()')[0]
        segmentation_key = device_info_root.xpath('./segmentation_supported/text()')[0]

        self.thisDevice = LocalDeviceObject(
            objectName=name_key,
            objectIdentifier=int(id_key),
            maxApduLengthAccepted=int(apdu_length_key),
            segmentationSupported=segmentation_key,
            vendorName=vendor_name_key,
            vendorIdentifier=int(vendor_identifier_key)
        )
        self.bacnet_app = None
        self.server = None  # Initialize later
        logger.info('Conpot Bacnet initialized using the %s template.', template)

    def handle(self, data, address):
        session = conpot_core.get_session('bacnet', address[0], address[1])
        logger.info('New Bacnet connection from %s:%d. (%s)', address[0], address[1], session.id)
        session.add_event({'type': 'NEW_CONNECTION'})
        # I'm not sure if gevent DatagramServer handles issues where the
        # received data is over the MTU -> fragmentation
        if data:
            pdu = PDU()
            pdu.pduData = bytearray(data)
            apdu = APDU()
            try:
                apdu.decode(pdu)
            except DecodingError:
                logger.warning("DecodingError - PDU: {}".format(pdu))
                return
            self.bacnet_app.indication(apdu, address, self.thisDevice)
            # send an appropriate response from BACnet app to the attacker
            self.bacnet_app.response(self.bacnet_app._response, address)
        logger.info('Bacnet client disconnected %s:%d. (%s)', address[0], address[1], session.id)

    def start(self, host, port):
        connection = (host, port)
        self.server = DatagramServer(connection, self.handle)
        # start to init the socket
        self.server.start()
        self.server.socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        # create application instance
        # not too beautiful, but the BACnetApp needs access to the socket's sendto method
        # this could properly be refactored in a way such that sending operates on it's own
        # (non-bound) socket.
        self.bacnet_app = BACnetApp(self.thisDevice, self.server)
        # get object_list and properties
        self.bacnet_app.get_objects_and_properties(self.dom)

        logger.info('Bacnet server started on: %s', connection)
        self.server.serve_forever()

    def stop(self):
        self.server.stop()


if __name__ == '__main__':
    import os
    test_template = os.getcwd() + '/../../templates/default/bacnet/bacnet.xml'
    test = BacnetServer(test_template, None, None)
    try:
        from bacpypes.apdu import WhoIsRequest, WhoHasObject, WhoHasRequest
        # code for generating adpu - who-is
        # request = WhoIsRequest(deviceInstanceRangeLowLimit=500, deviceInstanceRangeHighLimit=50000)
        # test_pdu = PDU()
        # test_apdu = APDU()
        # request.encode(test_apdu)
        # test_apdu.encode(test_pdu)
        # bacnet_app = BACnetApp(test.thisDevice, test)
        # bacnet_app.get_objects_and_properties(test.dom)
        # bacnet_app.indication(test_apdu, ('127.0.0.1', 9999), test.thisDevice)
        # print(bacnet_app._response)
        # bacnet_app.response(bacnet_app._response, ('127.0.0.1', 9999))
        # # logger.debug('Starting BACnet Server! at {}:{}'.format('localhost', 9999))
        # # test.start('127.0.0.1', 9999)

        # testing who-has
        request_object = WhoHasObject()
        request_object.objectIdentifier = ('binaryInput', 12)
        request = WhoHasRequest(object=request_object)
        test_apdu = APDU()
        request.encode(test_apdu)
        test_pdu = PDU()
        test_apdu.encode(test_pdu)
        bacnet_app = BACnetApp(test.thisDevice, test)
        bacnet_app.get_objects_and_properties(test.dom)
        bacnet_app.indication(test_apdu, ('127.0.0.1', 9999), test.thisDevice)
        print(bacnet_app._response)
        bacnet_app.response(bacnet_app._response, ('127.0.0.1', 9999))
    except KeyboardInterrupt:
        logger.debug('Stopping BACnet server')
        test.stop()